from .metaApiWebsocket_client import MetaApiWebsocketClient
from socketio import AsyncServer
from aiohttp import web
from ...metaApi.models import date, format_date
import pytest
import asyncio
import copy
import re
from urllib.parse import parse_qs
from mock import MagicMock, AsyncMock
from copy import deepcopy
from datetime import datetime
sio = None
client: MetaApiWebsocketClient = None


class FakeServer:

    def __init__(self):
        self.app = web.Application()
        self.runner = None

    async def start(self):
        port = 8080
        global sio
        sio = AsyncServer(async_mode='aiohttp')

        @sio.event
        async def connect(sid, environ):
            qs = parse_qs(environ['QUERY_STRING'])
            if ('auth-token' not in qs) or (qs['auth-token'] != ['token']):
                environ.emit({'error': 'UnauthorizedError', 'message': 'Authorization token invalid'})
                environ.close()

        sio.attach(self.app, socketio_path='ws')
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        site = web.TCPSite(self.runner, 'localhost', port)
        await site.start()

    async def stop(self):
        await self.runner.cleanup()


@pytest.fixture(autouse=True)
async def run_around_tests():
    FinalMock.__await__ = lambda x: async_magic_close().__await__()
    fake_server = FakeServer()
    await fake_server.start()
    global client
    client = MetaApiWebsocketClient('token', {'application': 'application', 'domain': 'project-stock.agiliumlabs.cloud',
                                              'requestTimeout': 3, 'retryOpts': {'retries': 3,
                                                                                 'minDelayInSeconds': 0.1,
                                                                                 'maxDelayInSeconds': 0.5}})
    client.set_url('http://localhost:8080')
    await client.connect()
    client._resolved = True
    yield
    await client.close()
    await fake_server.stop()
    tasks = [task for task in asyncio.all_tasks() if task is not
             asyncio.tasks.current_task()]
    list(map(lambda task: task.cancel(), tasks))


# This method closes the client once the required socket event has been called
async def async_magic_close():
    await client.close()


class FinalMock(MagicMock):
    def __init__(self, *args, **kwargs):
        super(MagicMock, self).__init__(*args, **kwargs)


FinalMock.__await__ = lambda x: async_magic_close().__await__()


class TestMetaApiWebsocketClient:
    @pytest.mark.asyncio
    async def test_retrieve_account(self):
        """Should retrieve MetaTrader account information from API."""

        account_information = {
            'broker': 'True ECN Trading Ltd',
            'currency': 'USD',
            'server': 'ICMarketsSC-Demo',
            'balance': 7319.9,
            'equity': 7306.649913200001,
            'margin': 184.1,
            'freeMargin': 7120.22,
            'leverage': 100,
            'marginLevel': 3967.58283542
        }

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'getAccountInformation' and data['accountId'] == 'accountId' \
                    and data['application'] == 'RPC':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'accountInformation': account_information})

        actual = await client.get_account_information('accountId')
        assert actual == account_information

    @pytest.mark.asyncio
    async def test_retrieve_positions(self):
        """Should retrieve MetaTrader positions from API."""

        positions = [{
            'id': '46214692',
            'type': 'POSITION_TYPE_BUY',
            'symbol': 'GBPUSD',
            'magic': 1000,
            'time': '2020-04-15T02:45:06.521Z',
            'updateTime': '2020-04-15T02:45:06.521Z',
            'openPrice': 1.26101,
            'currentPrice': 1.24883,
            'currentTickValue': 1,
            'volume': 0.07,
            'swap': 0,
            'profit': -85.25999999999966,
            'commission': -0.25,
            'clientId': 'TE_GBPUSD_7hyINWqAlE',
            'stopLoss': 1.17721,
            'unrealizedProfit': -85.25999999999901,
            'realizedProfit': -6.536993168992922e-13
        }]

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'getPositions' and data['accountId'] == 'accountId' \
                    and data['application'] == 'RPC':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'positions': positions})
            else:
                raise Exception('Wrong request')

        actual = await client.get_positions('accountId')
        positions[0]['time'] = date(positions[0]['time'])
        positions[0]['updateTime'] = date(positions[0]['updateTime'])
        assert actual == positions

    @pytest.mark.asyncio
    async def test_retrieve_position(self):
        """Should retrieve MetaTrader position from API."""

        position = {
            'id': '46214692',
            'type': 'POSITION_TYPE_BUY',
            'symbol': 'GBPUSD',
            'magic': 1000,
            'time': '2020-04-15T02:45:06.521Z',
            'updateTime': '2020-04-15T02:45:06.521Z',
            'openPrice': 1.26101,
            'currentPrice': 1.24883,
            'currentTickValue': 1,
            'volume': 0.07,
            'swap': 0,
            'profit': -85.25999999999966,
            'commission': -0.25,
            'clientId': 'TE_GBPUSD_7hyINWqAlE',
            'stopLoss': 1.17721,
            'unrealizedProfit': -85.25999999999901,
            'realizedProfit': -6.536993168992922e-13
        }

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'getPosition' and data['accountId'] == 'accountId' and data['positionId'] == '46214692' \
                    and data['application'] == 'RPC':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'position': position})

        actual = await client.get_position('accountId', '46214692')
        position['time'] = date(position['time'])
        position['updateTime'] = date(position['updateTime'])
        assert actual == position

    @pytest.mark.asyncio
    async def test_retrieve_orders(self):
        """Should retrieve MetaTrader orders from API."""

        orders = [{
            'id': '46871284',
            'type': 'ORDER_TYPE_BUY_LIMIT',
            'state': 'ORDER_STATE_PLACED',
            'symbol': 'AUDNZD',
            'magic': 123456,
            'platform': 'mt5',
            'time': '2020-04-20T08:38:58.270Z',
            'openPrice': 1.03,
            'currentPrice': 1.05206,
            'volume': 0.01,
            'currentVolume': 0.01,
            'comment': 'COMMENT2'
        }]

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'getOrders' and data['accountId'] == 'accountId' \
                    and data['application'] == 'RPC':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'orders': orders})

        actual = await client.get_orders('accountId')
        orders[0]['time'] = date(orders[0]['time'])
        assert actual == orders

    @pytest.mark.asyncio
    async def test_retrieve_order(self):
        """Should retrieve MetaTrader order from API by id."""

        order = {
            'id': '46871284',
            'type': 'ORDER_TYPE_BUY_LIMIT',
            'state': 'ORDER_STATE_PLACED',
            'symbol': 'AUDNZD',
            'magic': 123456,
            'platform': 'mt5',
            'time': '2020-04-20T08:38:58.270Z',
            'openPrice': 1.03,
            'currentPrice': 1.05206,
            'volume': 0.01,
            'currentVolume': 0.01,
            'comment': 'COMMENT2'
        }

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'getOrder' and data['accountId'] == 'accountId' and data['orderId'] == '46871284' \
                    and data['application'] == 'RPC':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'order': order})

        actual = await client.get_order('accountId', '46871284')
        order['time'] = date(order['time'])
        assert actual == order

    @pytest.mark.asyncio
    async def test_retrieve_history_orders_by_ticket(self):
        """Should retrieve MetaTrader history orders from API by ticket."""

        history_orders = [{
            'clientId': 'TE_GBPUSD_7hyINWqAlE',
            'currentPrice': 1.261,
            'currentVolume': 0,
            'doneTime': '2020-04-15T02:45:06.521Z',
            'id': '46214692',
            'magic': 1000,
            'platform': 'mt5',
            'positionId': '46214692',
            'state': 'ORDER_STATE_FILLED',
            'symbol': 'GBPUSD',
            'time': '2020-04-15T02:45:06.260Z',
            'type': 'ORDER_TYPE_BUY',
            'volume': 0.07
        }]

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'getHistoryOrdersByTicket' and data['accountId'] == 'accountId' and \
                    data['ticket'] == '46214692' and data['application'] == 'RPC':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'historyOrders': history_orders,
                                            'synchronizing': False})

        actual = await client.get_history_orders_by_ticket('accountId', '46214692')
        history_orders[0]['time'] = date(history_orders[0]['time'])
        history_orders[0]['doneTime'] = date(history_orders[0]['doneTime'])
        assert actual == {'historyOrders': history_orders, 'synchronizing': False}

    @pytest.mark.asyncio
    async def test_retrieve_history_orders_by_position(self):
        """Should retrieve MetaTrader history orders from API by position."""

        history_orders = [{
            'clientId': 'TE_GBPUSD_7hyINWqAlE',
            'currentPrice': 1.261,
            'currentVolume': 0,
            'doneTime': '2020-04-15T02:45:06.521Z',
            'id': '46214692',
            'magic': 1000,
            'platform': 'mt5',
            'positionId': '46214692',
            'state': 'ORDER_STATE_FILLED',
            'symbol': 'GBPUSD',
            'time': '2020-04-15T02:45:06.260Z',
            'type': 'ORDER_TYPE_BUY',
            'volume': 0.07
        }]

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'getHistoryOrdersByPosition' and data['accountId'] == 'accountId' and \
                    data['positionId'] == '46214692' and data['application'] == 'RPC':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'historyOrders': history_orders,
                                            'synchronizing': False})

        actual = await client.get_history_orders_by_position('accountId', '46214692')
        history_orders[0]['time'] = date(history_orders[0]['time'])
        history_orders[0]['doneTime'] = date(history_orders[0]['doneTime'])
        assert actual == {'historyOrders': history_orders, 'synchronizing': False}

    @pytest.mark.asyncio
    async def test_retrieve_history_orders_by_time_range(self):
        """Should retrieve MetaTrader history orders from API by time range."""

        history_orders = [{
            'clientId': 'TE_GBPUSD_7hyINWqAlE',
            'currentPrice': 1.261,
            'currentVolume': 0,
            'doneTime': '2020-04-15T02:45:06.521Z',
            'id': '46214692',
            'magic': 1000,
            'platform': 'mt5',
            'positionId': '46214692',
            'state': 'ORDER_STATE_FILLED',
            'symbol': 'GBPUSD',
            'time': '2020-04-15T02:45:06.260Z',
            'type': 'ORDER_TYPE_BUY',
            'volume': 0.07
        }]

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'getHistoryOrdersByTimeRange' and data['accountId'] == 'accountId' and \
                    data['startTime'] == '2020-04-15T02:45:00.000Z' and data['application'] == 'RPC' and \
                    data['endTime'] == '2020-04-15T02:46:00.000Z' and data['offset'] == 1 and data['limit'] == 100:
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'historyOrders': history_orders,
                                            'synchronizing': False})

        actual = await client.get_history_orders_by_time_range('accountId', date('2020-04-15T02:45:00.000Z'),
                                                               date('2020-04-15T02:46:00.000Z'), 1, 100)
        history_orders[0]['time'] = date(history_orders[0]['time'])
        history_orders[0]['doneTime'] = date(history_orders[0]['doneTime'])
        assert actual == {'historyOrders': history_orders, 'synchronizing': False}

    @pytest.mark.asyncio
    async def test_retrieve_deals_by_ticket(self):
        """Should retrieve MetaTrader deals from API by ticket."""

        deals = [{
            'clientId': 'TE_GBPUSD_7hyINWqAlE',
            'commission': -0.25,
            'entryType': 'DEAL_ENTRY_IN',
            'id': '33230099',
            'magic': 1000,
            'platform': 'mt5',
            'orderId': '46214692',
            'positionId': '46214692',
            'price': 1.26101,
            'profit': 0,
            'swap': 0,
            'symbol': 'GBPUSD',
            'time': '2020-04-15T02:45:06.521Z',
            'type': 'DEAL_TYPE_BUY',
            'volume': 0.07
        }]

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'getDealsByTicket' and data['accountId'] == 'accountId' and \
                    data['ticket'] == '46214692' and data['application'] == 'RPC':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'deals': deals,
                                            'synchronizing': False})

        actual = await client.get_deals_by_ticket('accountId', '46214692')
        deals[0]['time'] = date(deals[0]['time'])
        assert actual == {'deals': deals, 'synchronizing': False}

    @pytest.mark.asyncio
    async def test_retrieve_deals_by_position(self):
        """Should retrieve MetaTrader deals from API by position."""

        deals = [{
            'clientId': 'TE_GBPUSD_7hyINWqAlE',
            'commission': -0.25,
            'entryType': 'DEAL_ENTRY_IN',
            'id': '33230099',
            'magic': 1000,
            'platform': 'mt5',
            'orderId': '46214692',
            'positionId': '46214692',
            'price': 1.26101,
            'profit': 0,
            'swap': 0,
            'symbol': 'GBPUSD',
            'time': '2020-04-15T02:45:06.521Z',
            'type': 'DEAL_TYPE_BUY',
            'volume': 0.07
        }]

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'getDealsByPosition' and data['accountId'] == 'accountId' and \
                    data['positionId'] == '46214692' and data['application'] == 'RPC':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'deals': deals,
                                            'synchronizing': False})

        actual = await client.get_deals_by_position('accountId', '46214692')
        deals[0]['time'] = date(deals[0]['time'])
        assert actual == {'deals': deals, 'synchronizing': False}

    @pytest.mark.asyncio
    async def test_retrieve_deals_by_time_range(self):
        """Should retrieve MetaTrader deals from API by time range."""

        deals = [{
            'clientId': 'TE_GBPUSD_7hyINWqAlE',
            'commission': -0.25,
            'entryType': 'DEAL_ENTRY_IN',
            'id': '33230099',
            'magic': 1000,
            'platform': 'mt5',
            'orderId': '46214692',
            'positionId': '46214692',
            'price': 1.26101,
            'profit': 0,
            'swap': 0,
            'symbol': 'GBPUSD',
            'time': '2020-04-15T02:45:06.521Z',
            'type': 'DEAL_TYPE_BUY',
            'volume': 0.07
        }]

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'getDealsByTimeRange' and data['accountId'] == 'accountId' and \
                    data['startTime'] == '2020-04-15T02:45:00.000Z' and data['application'] == 'RPC' and \
                    data['endTime'] == '2020-04-15T02:46:00.000Z' and data['offset'] == 1 and data['limit'] == 100:
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'deals': deals,
                                            'synchronizing': False})

        actual = await client.get_deals_by_time_range('accountId', date('2020-04-15T02:45:00.000Z'),
                                                      date('2020-04-15T02:46:00.000Z'), 1, 100)
        deals[0]['time'] = date(deals[0]['time'])
        assert actual == {'deals': deals, 'synchronizing': False}

    @pytest.mark.asyncio
    async def test_remove_history(self):
        """Should remove history from API."""

        request_received = False

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'removeHistory' and data['accountId'] == 'accountId' \
                    and data['application'] == 'app':
                nonlocal request_received
                request_received = True
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId']})

        await client.remove_history('accountId', 'app')
        assert request_received

    @pytest.mark.asyncio
    async def test_remove_application(self):
        """Should remove application from API."""

        request_received = False

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'removeApplication' and data['accountId'] == 'accountId' \
                    and data['application'] == 'application':
                nonlocal request_received
                request_received = True
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId']})

        await client.remove_application('accountId')
        assert request_received

    @pytest.mark.asyncio
    async def test_execute_trade(self):
        """Should execute a trade via new API version."""

        trade = {
            'actionType': 'ORDER_TYPE_SELL',
            'symbol': 'AUDNZD',
            'volume': 0.07
        }
        response = {
            'numericCode': 10009,
            'stringCode': 'TRADE_RETCODE_DONE',
            'message': 'Request completed',
            'orderId': '46870472'
        }

        @sio.on('request')
        async def on_request(sid, data):
            assert data['trade'] == trade
            if data['type'] == 'trade' and data['accountId'] == 'accountId' and data['application'] == 'application':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'response': response})

        actual = await client.trade('accountId', trade)
        assert actual == response

    @pytest.mark.asyncio
    async def test_fail_trade_on_old_api(self):
        """Should execute a trade via API and receive trade error from old API version."""

        trade = {
            'actionType': 'ORDER_TYPE_SELL',
            'symbol': 'AUDNZD',
            'volume': 0.07
        }
        response = {
            'error': 10006,
            'description': 'TRADE_RETCODE_REJECT',
            'message': 'Request rejected',
            'orderId': '46870472'
        }

        @sio.on('request')
        async def on_request(sid, data):
            assert data['trade'] == trade
            if data['type'] == 'trade' and data['accountId'] == 'accountId' and data['application'] == 'application':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'response': response})

        try:
            await client.trade('accountId', trade)
            Exception('TradeException expected')
        except Exception as err:
            assert err.__class__.__name__ == 'TradeException'
            assert err.__str__() == 'Request rejected'
            assert err.stringCode == 'TRADE_RETCODE_REJECT'
            assert err.numericCode == 10006

    @pytest.mark.asyncio
    async def test_connect_to_terminal(self):
        """Should connect to MetaTrader terminal."""
        request_received = False

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'subscribe' and data['accountId'] == 'accountId' \
                    and data['application'] == 'application' and data['instanceIndex'] == 1:
                nonlocal request_received
                request_received = True
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId']})

        await client.subscribe('accountId', 1)
        await asyncio.sleep(0.05)
        assert request_received

    @pytest.mark.asyncio
    async def test_return_error_if_failed(self):
        """Should return error if connect to MetaTrader terminal failed."""
        request_received = False

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'subscribe' and data['accountId'] == 'accountId' \
                    and data['application'] == 'application':
                nonlocal request_received
                request_received = True
            await sio.emit('processingError', {'id': 1, 'error': 'NotAuthenticatedError', 'message': 'Error message',
                                               'requestId': data['requestId']})

        success = True
        try:
            await client.subscribe('accountId')
            await asyncio.sleep(0.05)
            success = False
        except Exception as err:
            assert err.__class__.__name__ == 'NotConnectedException'
        assert success
        assert request_received

    @pytest.mark.asyncio
    async def test_reconnect_to_terminal(self):
        """Should reconnect to MetaTrader terminal."""

        request_received = False

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'reconnect' and data['accountId'] == 'accountId' \
                    and data['application'] == 'application':
                nonlocal request_received
                request_received = True
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId']})

        await client.reconnect('accountId')
        assert request_received

    @pytest.mark.asyncio
    async def test_retrieve_symbol_specification(self):
        """Should retrieve symbol specification from API."""

        specification = {
            'symbol': 'AUDNZD',
            'tickSize': 0.00001,
            'minVolume': 0.01,
            'maxVolume': 100,
            'volumeStep': 0.01
        }

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'getSymbolSpecification' and data['accountId'] == 'accountId' and \
                    data['symbol'] == 'AUDNZD' and data['application'] == 'RPC':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'specification': specification})

        actual = await client.get_symbol_specification('accountId', 'AUDNZD')
        assert actual == specification

    @pytest.mark.asyncio
    async def test_retrieve_symbol_price(self):
        """Should retrieve symbol price from API."""

        price = {
            'symbol': 'AUDNZD',
            'bid': 1.05297,
            'ask': 1.05309,
            'profitTickValue': 0.59731,
            'lossTickValue': 0.59736
        }

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'getSymbolPrice' and data['accountId'] == 'accountId' and \
                    data['symbol'] == 'AUDNZD' and data['application'] == 'RPC':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'price': price})

        actual = await client.get_symbol_price('accountId', 'AUDNZD')
        assert actual == price

    @pytest.mark.asyncio
    async def test_send_uptime_stats(self):
        """Should send uptime stats to the server."""

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'saveUptime' and data['accountId'] == 'accountId' and \
                    data['uptime'] == {'1h': 100} and data['application'] == 'application':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId']})
        await client.save_uptime('accountId', {'1h': 100})

    @pytest.mark.asyncio
    async def test_unsubscribe(self):
        """Should unsubscribe from account data."""

        response = {'type': 'response', 'accountId': 'accountId'}

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'unsubscribe' and data['accountId'] == 'accountId':
                await sio.emit('response', {'requestId': data['requestId'], **response})

        actual = await client.unsubscribe('accountId')
        assert actual['type'] == response['type']
        assert actual['accountId'] == response['accountId']

    @pytest.mark.asyncio
    async def test_handle_validation_exception(self):
        """Should handle ValidationError."""

        trade = {
            'actionType': 'ORDER_TYPE_SELL',
            'symbol': 'AUDNZD'
        }

        @sio.on('request')
        async def on_request(sid, data):
            await sio.emit('processingError', {'id': 1, 'error': 'ValidationError', 'message': 'Validation failed',
                           'details': [{'parameter': 'volume', 'message': 'Required value.'}],
                                               'requestId': data['requestId']})

        try:
            await client.trade('accountId', trade)
            Exception('ValidationError expected')
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'

    @pytest.mark.asyncio
    async def test_handle_not_found_exception(self):
        """Should handle NotFoundError."""

        @sio.on('request')
        async def on_request(sid, data):
            await sio.emit('processingError',  {'id': 1, 'error': 'NotFoundError',
                                                'message': 'Position id 1234 not found',
                                                'requestId': data['requestId']})

        try:
            await client.get_position('accountId', '1234')
            Exception('NotFoundException expected')
        except Exception as err:
            assert err.__class__.__name__ == 'NotFoundException'

    @pytest.mark.asyncio
    async def test_handle_not_synchronized_exception(self):
        """Should handle NotSynchronizedError."""

        @sio.on('request')
        async def on_request(sid, data):
            await sio.emit('processingError', {'id': 1, 'error': 'NotSynchronizedError', 'message': 'Error message',
                                               'requestId': data['requestId']})

        try:
            await client.get_position('accountId', '1234')
            Exception('NotSynchronizedError expected')
        except Exception as err:
            assert err.__class__.__name__ == 'NotSynchronizedException'

    @pytest.mark.asyncio
    async def test_handle_not_connected_exception(self):
        """Should handle NotSynchronizedError."""

        @sio.on('request')
        async def on_request(sid, data):
            await sio.emit('processingError', {'id': 1, 'error': 'NotAuthenticatedError', 'message': 'Error message',
                                               'requestId': data['requestId']})

        try:
            await client.get_position('accountId', '1234')
            Exception('NotConnectedError expected')
        except Exception as err:
            assert err.__class__.__name__ == 'NotConnectedException'

    @pytest.mark.asyncio
    async def test_handle_other_exceptions(self):
        """Should handle other errors."""

        @sio.on('request')
        async def on_request(sid, data):
            await sio.emit('processingError', {'id': 1, 'error': 'Error', 'message': 'Error message',
                                               'requestId': data['requestId']})

        try:
            await client.get_position('accountId', '1234')
            Exception('InternalError expected')
        except Exception as err:
            assert err.__class__.__name__ == 'InternalException'

    @pytest.mark.asyncio
    async def test_process_auth_sync_event(self):
        """Should process authenticated synchronization event."""
        listener = MagicMock()
        listener.on_connected = FinalMock()
        client.add_synchronization_listener('accountId', listener)
        await sio.emit('synchronization', {'type': 'authenticated', 'accountId': 'accountId', 'instanceIndex': 1,
                                           'replicas': 2})
        await client._socket.wait()
        listener.on_connected.assert_called_with(1, 2)

    @pytest.mark.asyncio
    async def test_process_auth_sync_event_with_session_id(self):
        """Should process authenticated synchronization event with session id."""
        listener = MagicMock()
        listener.on_connected = FinalMock()
        client.add_synchronization_listener('accountId', listener)
        await sio.emit('synchronization', {'type': 'authenticated', 'accountId': 'accountId', 'instanceIndex': 2,
                                           'replicas': 4, 'sessionId': 'wrong'})
        await sio.emit('synchronization', {'type': 'authenticated', 'accountId': 'accountId', 'instanceIndex': 1,
                                           'replicas': 2, 'sessionId': client._sessionId})
        await client._socket.wait()
        assert listener.on_connected.call_count == 1
        listener.on_connected.assert_called_with(1, 2)

    @pytest.mark.asyncio
    async def test_process_broker_connection_status_event(self):
        """Should process broker connection status event."""
        listener = MagicMock()
        listener.on_broker_connection_status_changed = FinalMock()
        client.add_synchronization_listener('accountId', listener)
        await sio.emit('synchronization', {'type': 'authenticated', 'accountId': 'accountId', 'host': 'ps-mpa-1',
                                           'instanceIndex': 1})
        await sio.emit('synchronization', {'type': 'status', 'accountId': 'accountId', 'host': 'ps-mpa-1',
                                           'connected': True, 'instanceIndex': 1})
        await client._socket.wait()
        listener.on_broker_connection_status_changed.assert_called_with(1, True)

    @pytest.mark.asyncio
    async def test_process_server_health_status(self):
        """Should process server-side health status event."""
        listener = MagicMock()
        listener.on_broker_connection_status_changed = AsyncMock()
        listener.on_health_status = FinalMock()
        client.add_synchronization_listener('accountId', listener)
        await sio.emit('synchronization', {'type': 'authenticated', 'accountId': 'accountId', 'host': 'ps-mpa-1',
                                           'instanceIndex': 1})
        await sio.emit('synchronization', {'type': 'status', 'accountId': 'accountId', 'host': 'ps-mpa-1',
                                           'connected': True, 'healthStatus': {'restApiHealthy': True},
                                           'instanceIndex': 1})
        await client._socket.wait()
        listener.on_health_status.assert_called_with(1, {'restApiHealthy': True})

    @pytest.mark.asyncio
    async def test_process_disconnected_synchronization_event(self):
        """Should process disconnected synchronization event."""

        listener = MagicMock()
        listener.on_disconnected = FinalMock()
        client.add_synchronization_listener('accountId', listener)
        await sio.emit('synchronization', {'type': 'authenticated', 'accountId': 'accountId', 'host': 'ps-mpa-1',
                                           'instanceIndex': 1})
        await sio.emit('synchronization', {'type': 'disconnected', 'accountId': 'accountId', 'host': 'ps-mpa-1',
                                           'instanceIndex': 1})
        await client._socket.wait()
        listener.on_disconnected.assert_called_with(1)

    @pytest.mark.asyncio
    async def test_synchronize_with_metatrader_terminal(self):
        """Should synchronize with MetaTrader terminal."""

        request_received = False

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'synchronize' and data['accountId'] == 'accountId' and \
                    data['startingHistoryOrderTime'] == '2020-01-01T00:00:00.000Z' and \
                    data['startingDealTime'] == '2020-01-02T00:00:00.000Z' and \
                    data['requestId'] == 'synchronizationId' and data['application'] == 'application' and \
                    data['instanceIndex'] == 1:
                nonlocal request_received
                request_received = True
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId']})

        await client.synchronize('accountId', 1, 'synchronizationId', date('2020-01-01T00:00:00.000Z'),
                                 date('2020-01-02T00:00:00.000Z'))
        assert request_received

    @pytest.mark.asyncio
    async def test_process_sync_started(self):
        """Should process synchronization started event."""
        listener = MagicMock()
        listener.on_synchronization_started = FinalMock()

        client.add_synchronization_listener('accountId', listener)
        await sio.emit('synchronization', {'type': 'synchronizationStarted', 'accountId': 'accountId',
                                           'instanceIndex': 1})
        await client._socket.wait()
        listener.on_synchronization_started.assert_called_with(1)

    @pytest.mark.asyncio
    async def test_synchronize_account_information(self):
        """Should synchronize account information."""

        account_information = {
            'broker': 'True ECN Trading Ltd',
            'currency': 'USD',
            'server': 'ICMarketsSC-Demo',
            'balance': 7319.9,
            'equity': 7306.649913200001,
            'margin': 184.1,
            'freeMargin': 7120.22,
            'leverage': 100,
            'marginLevel': 3967.58283542
        }
        listener = MagicMock()
        listener.on_account_information_updated = FinalMock()

        client.add_synchronization_listener('accountId', listener)
        await sio.emit('synchronization', {'type': 'accountInformation', 'accountId': 'accountId',
                                           'accountInformation': account_information, 'instanceIndex': 1})
        await client._socket.wait()
        listener.on_account_information_updated.assert_called_with(1, account_information)

    @pytest.mark.asyncio
    async def test_synchronize_positions(self):
        """Should synchronize positions."""

        positions = [{
            'id': '46214692',
            'type': 'POSITION_TYPE_BUY',
            'symbol': 'GBPUSD',
            'magic': 1000,
            'time': '2020-04-15T02:45:06.521Z',
            'updateTime': '2020-04-15T02:45:06.521Z',
            'openPrice': 1.26101,
            'currentPrice': 1.24883,
            'currentTickValue': 1,
            'volume': 0.07,
            'swap': 0,
            'profit': -85.25999999999966,
            'commission': -0.25,
            'clientId': 'TE_GBPUSD_7hyINWqAlE',
            'stopLoss': 1.17721,
            'unrealizedProfit': -85.25999999999901,
            'realizedProfit': -6.536993168992922e-13
        }]
        listener = MagicMock()
        listener.on_positions_replaced = FinalMock()

        client.add_synchronization_listener('accountId', listener)
        await sio.emit('synchronization', {'type': 'positions', 'accountId': 'accountId', 'positions': positions,
                                           'instanceIndex': 1})
        await client._socket.wait()
        positions[0]['time'] = date(positions[0]['time'])
        positions[0]['updateTime'] = date(positions[0]['updateTime'])
        listener.on_positions_replaced.assert_called_with(1, positions)

    @pytest.mark.asyncio
    async def test_synchronize_orders(self):
        """Should synchronize orders."""

        orders = [{
            'id': '46871284',
            'type': 'ORDER_TYPE_BUY_LIMIT',
            'state': 'ORDER_STATE_PLACED',
            'symbol': 'AUDNZD',
            'magic': 123456,
            'platform': 'mt5',
            'time': '2020-04-20T08:38:58.270Z',
            'openPrice': 1.03,
            'currentPrice': 1.05206,
            'volume': 0.01,
            'currentVolume': 0.01,
            'comment': 'COMMENT2'
        }]
        listener = MagicMock()
        listener.on_orders_replaced = FinalMock()

        client.add_synchronization_listener('accountId', listener)
        await sio.emit('synchronization', {'type': 'orders', 'accountId': 'accountId', 'orders': orders,
                                           'instanceIndex': 1})
        await client._socket.wait()
        orders[0]['time'] = date(orders[0]['time'])
        listener.on_orders_replaced.assert_called_with(1, orders)

    @pytest.mark.asyncio
    async def test_synchronize_history_orders(self):
        """Should synchronize history orders."""

        history_orders = [{
            'clientId': 'TE_GBPUSD_7hyINWqAlE',
            'currentPrice': 1.261,
            'currentVolume': 0,
            'doneTime': '2020-04-15T02:45:06.521Z',
            'id': '46214692',
            'magic': 1000,
            'platform': 'mt5',
            'positionId': '46214692',
            'state': 'ORDER_STATE_FILLED',
            'symbol': 'GBPUSD',
            'time': '2020-04-15T02:45:06.260Z',
            'type': 'ORDER_TYPE_BUY',
            'volume': 0.07
        }]
        listener = MagicMock()
        listener.on_history_order_added = FinalMock()
        client.add_synchronization_listener('accountId', listener)
        await sio.emit('synchronization', {'type': 'historyOrders', 'accountId': 'accountId',
                                           'historyOrders': history_orders, 'instanceIndex': 1})
        await client._socket.wait()
        history_orders[0]['time'] = date(history_orders[0]['time'])
        history_orders[0]['doneTime'] = date(history_orders[0]['doneTime'])
        listener.on_history_order_added.assert_called_with(1, history_orders[0])

    @pytest.mark.asyncio
    async def test_synchronize_deals(self):
        """Should synchronize deals."""

        deals = [{
            'clientId': 'TE_GBPUSD_7hyINWqAlE',
            'commission': -0.25,
            'entryType': 'DEAL_ENTRY_IN',
            'id': '33230099',
            'magic': 1000,
            'platform': 'mt5',
            'orderId': '46214692',
            'positionId': '46214692',
            'price': 1.26101,
            'profit': 0,
            'swap': 0,
            'symbol': 'GBPUSD',
            'time': '2020-04-15T02:45:06.521Z',
            'type': 'DEAL_TYPE_BUY',
            'volume': 0.07
        }]
        listener = MagicMock()
        listener.on_deal_added = FinalMock()
        client.add_synchronization_listener('accountId', listener)
        await sio.emit('synchronization', {'type': 'deals', 'accountId': 'accountId', 'deals': deals,
                                           'instanceIndex': 1})
        await client._socket.wait()
        deals[0]['time'] = date(deals[0]['time'])
        listener.on_deal_added.assert_called_with(1, deals[0])

    @pytest.mark.asyncio
    async def test_process_synchronization_updates(self):
        """Should process synchronization updates."""

        update = {
            'accountInformation': {
                'broker': 'True ECN Trading Ltd',
                'currency': 'USD',
                'server': 'ICMarketsSC-Demo',
                'balance': 7319.9,
                'equity': 7306.649913200001,
                'margin': 184.1,
                'freeMargin': 7120.22,
                'leverage': 100,
                'marginLevel': 3967.58283542
            },
            'updatedPositions': [{
                'id': '46214692',
                'type': 'POSITION_TYPE_BUY',
                'symbol': 'GBPUSD',
                'magic': 1000,
                'time': '2020-04-15T02:45:06.521Z',
                'updateTime': '2020-04-15T02:45:06.521Z',
                'openPrice': 1.26101,
                'currentPrice': 1.24883,
                'currentTickValue': 1,
                'volume': 0.07,
                'swap': 0,
                'profit': -85.25999999999966,
                'commission': -0.25,
                'clientId': 'TE_GBPUSD_7hyINWqAlE',
                'stopLoss': 1.17721,
                'unrealizedProfit': -85.25999999999901,
                'realizedProfit': -6.536993168992922e-13
            }],
            'removedPositionIds': ['1234'],
            'updatedOrders': [{
                'id': '46871284',
                'type': 'ORDER_TYPE_BUY_LIMIT',
                'state': 'ORDER_STATE_PLACED',
                'symbol': 'AUDNZD',
                'magic': 123456,
                'platform': 'mt5',
                'time': '2020-04-20T08:38:58.270Z',
                'openPrice': 1.03,
                'currentPrice': 1.05206,
                'volume': 0.01,
                'currentVolume': 0.01,
                'comment': 'COMMENT2'
            }],
            'completedOrderIds': ['2345'],
            'historyOrders': [{
                'clientId': 'TE_GBPUSD_7hyINWqAlE',
                'currentPrice': 1.261,
                'currentVolume': 0,
                'doneTime': '2020-04-15T02:45:06.521Z',
                'id': '46214692',
                'magic': 1000,
                'platform': 'mt5',
                'positionId': '46214692',
                'state': 'ORDER_STATE_FILLED',
                'symbol': 'GBPUSD',
                'time': '2020-04-15T02:45:06.260Z',
                'type': 'ORDER_TYPE_BUY',
                'volume': 0.07
            }],
            'deals': [{
                'clientId': 'TE_GBPUSD_7hyINWqAlE',
                'commission': -0.25,
                'entryType': 'DEAL_ENTRY_IN',
                'id': '33230099',
                'magic': 1000,
                'platform': 'mt5',
                'orderId': '46214692',
                'positionId': '46214692',
                'price': 1.26101,
                'profit': 0,
                'swap': 0,
                'symbol': 'GBPUSD',
                'time': '2020-04-15T02:45:06.521Z',
                'type': 'DEAL_TYPE_BUY',
                'volume': 0.07
            }]
        }
        listener = MagicMock()
        listener.on_account_information_updated = AsyncMock()
        listener.on_position_updated = AsyncMock()
        listener.on_position_removed = AsyncMock()
        listener.on_order_updated = AsyncMock()
        listener.on_order_completed = AsyncMock()
        listener.on_history_order_added = AsyncMock()
        listener.on_deal_added = FinalMock()
        client.add_synchronization_listener('accountId', listener)
        emit = copy.deepcopy(update)
        emit['type'] = 'update'
        emit['accountId'] = 'accountId'
        emit['instanceIndex'] = 1
        await sio.emit('synchronization', emit)
        await client._socket.wait()
        update['updatedPositions'][0]['time'] = date(update['updatedPositions'][0]['time'])
        update['updatedPositions'][0]['updateTime'] = date(update['updatedPositions'][0]['updateTime'])
        update['updatedOrders'][0]['time'] = date(update['updatedOrders'][0]['time'])
        update['historyOrders'][0]['time'] = date(update['historyOrders'][0]['time'])
        update['historyOrders'][0]['doneTime'] = date(update['historyOrders'][0]['doneTime'])
        update['deals'][0]['time'] = date(update['deals'][0]['time'])
        listener.on_account_information_updated.assert_called_with(1, update['accountInformation'])
        listener.on_position_updated.assert_called_with(1, update['updatedPositions'][0])
        listener.on_position_removed.assert_called_with(1, update['removedPositionIds'][0])
        listener.on_order_updated.assert_called_with(1, update['updatedOrders'][0])
        listener.on_order_completed.assert_called_with(1, update['completedOrderIds'][0])
        listener.on_history_order_added.assert_called_with(1, update['historyOrders'][0])
        listener.on_deal_added.assert_called_with(1, update['deals'][0])

    @pytest.mark.asyncio
    async def test_retry_on_failure(self):
        """Should retry request on failure."""
        request_counter = 0
        order = {
            'id': '46871284',
            'type': 'ORDER_TYPE_BUY_LIMIT',
            'state': 'ORDER_STATE_PLACED',
            'symbol': 'AUDNZD',
            'magic': 123456,
            'platform': 'mt5',
            'time': '2020-04-20T08:38:58.270Z',
            'openPrice': 1.03,
            'currentPrice': 1.05206,
            'volume': 0.01,
            'currentVolume': 0.01,
            'comment': 'COMMENT2'
        }

        @sio.on('request')
        async def on_request(sid, data):
            nonlocal request_counter
            if request_counter > 1 and data['type'] == 'getOrder' and data['accountId'] == 'accountId' and \
                    data['orderId'] == '46871284' and data['application'] == 'RPC':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'order': order})
            request_counter += 1

        actual = await client.get_order('accountId', '46871284')
        order['time'] = date(order['time'])
        assert actual == order

    @pytest.mark.asyncio
    async def test_not_retry_on_failure(self):
        """Should not retry request on validation error."""
        request_counter = 0

        @sio.on('request')
        async def on_request(sid, data):
            nonlocal request_counter
            if request_counter > 0 and data['type'] == 'subscribeToMarketData' and data['accountId'] == 'accountId' \
                    and data['symbol'] == 'EURUSD' and data['application'] == 'application' and \
                    data['instanceIndex'] == 1:
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId']})
            else:
                await sio.emit('processingError', {'id': 1, 'error': 'ValidationError', 'message': 'Validation failed',
                                                   'details': [{'parameter': 'volume', 'message': 'Required value.'}],
                                                   'requestId': data['requestId']})
            request_counter += 1
            try:
                await client.subscribe_to_market_data('accountId', 1, 'EURUSD')
                Exception('ValidationException expected')
            except Exception as err:
                assert err.__class__.__name__ == 'ValidationException'
                await client.close()
            assert request_counter == 1

    @pytest.mark.asyncio
    async def test_not_retry_trade(self):
        """Should not retry trade requests on fail."""
        request_counter = 0
        trade = {
            'actionType': 'ORDER_TYPE_SELL',
            'symbol': 'AUDNZD',
            'volume': 0.07
        }

        @sio.on('request')
        async def on_request(sid, data):
            nonlocal request_counter
            if request_counter > 0:
                pytest.fail()
            request_counter += 1

        try:
            await client.trade('accountId', trade)
            Exception('TimeoutException expected')
        except Exception as err:
            assert err.__class__.__name__ == 'TimeoutException'
            await client.close()

    @pytest.mark.asyncio
    async def test_timeout_on_no_response(self):
        """Should return timeout error if no server response received."""

        trade = {
            'actionType': 'ORDER_TYPE_SELL',
            'symbol': 'AUDNZD',
            'volume': 0.07
        }

        @sio.on('request')
        async def on_request(sid, data):
            pass

        try:
            await client.trade('accountId', trade)
            Exception('TimeoutException expected')
        except Exception as err:
            assert err.__class__.__name__ == 'TimeoutException'
            await client.close()

    @pytest.mark.asyncio
    async def test_subscribe_to_market_data_with_mt_terminal(self):
        """Should subscribe to market data with MetaTrader terminal."""

        request_received = False

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'subscribeToMarketData' and data['accountId'] == 'accountId' and \
                    data['symbol'] == 'EURUSD' and data['application'] == 'application' and data['instanceIndex'] == 1:
                nonlocal request_received
                request_received = True
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId']})

        await client.subscribe_to_market_data('accountId', 1, 'EURUSD')
        assert request_received

    @pytest.mark.asyncio
    async def test_unsubscribe_from_market_data_with_mt_terminal(self):
        """Should unsubscribe from market data with MetaTrader terminal."""

        request_received = False

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'unsubscribeFromMarketData' and data['accountId'] == 'accountId' and \
                    data['symbol'] == 'EURUSD' and data['application'] == 'application' and data['instanceIndex'] == 1:
                nonlocal request_received
                request_received = True
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId']})

        await client.unsubscribe_from_market_data('accountId', 1, 'EURUSD')
        assert request_received

    @pytest.mark.asyncio
    async def test_synchronize_symbol_specifications(self):
        """Should synchronize symbol specifications."""

        specifications = [{
            'symbol': 'EURUSD',
            'tickSize': 0.00001,
            'minVolume': 0.01,
            'maxVolume': 200,
            'volumeStep': 0.01
        }]
        listener = MagicMock()
        listener.on_symbol_specification_updated = FinalMock()
        client.add_synchronization_listener('accountId', listener)
        await sio.emit('synchronization', {'type': 'specifications', 'accountId': 'accountId',
                                           'specifications': specifications, 'instanceIndex': 1})
        await client._socket.wait()
        listener.on_symbol_specification_updated.assert_called_with(1, specifications[0])

    @pytest.mark.asyncio
    async def test_synchronize_symbol_prices(self):
        """Should synchronize symbol prices."""

        prices = [{
            'symbol': 'AUDNZD',
            'bid': 1.05916,
            'ask': 1.05927,
            'profitTickValue': 0.602,
            'lossTickValue': 0.60203
        }]
        listener = MagicMock()
        listener.on_symbol_prices_updated = AsyncMock()
        listener.on_symbol_price_updated = FinalMock()
        client.add_synchronization_listener('accountId', listener)
        await sio.emit('synchronization', {'type': 'prices', 'accountId': 'accountId', 'prices': prices,
                                           'equity': 100, 'margin': 200, 'freeMargin': 400, 'marginLevel': 40000,
                                           'instanceIndex': 1})
        await client._socket.wait()
        listener.on_symbol_prices_updated.assert_called_with(1, prices, 100, 200, 400, 40000)
        listener.on_symbol_price_updated.assert_called_with(1, prices[0])

    @pytest.mark.asyncio
    async def test_wait_for_server_side_sync(self):
        """Should wait for server-side terminal state synchronization."""

        request_received = False

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'waitSynchronized' and data['accountId'] == 'accountId' and \
                    data['applicationPattern'] == 'app.*' and data['timeoutInSeconds'] == 10 \
                    and data['application'] == 'application' and data['instanceIndex'] == 1:
                nonlocal request_received
                request_received = True
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId']})

        await client.wait_synchronized('accountId', 1, 'app.*', 10)
        assert request_received

    @pytest.mark.asyncio
    async def test_invoke_latency_listener(self):
        """Should invoke latency listener on response."""

        account_id = None
        request_type = None
        actual_timestamps = None

        def on_response(aid, type, ts):
            nonlocal account_id
            account_id = aid
            nonlocal request_type
            request_type = type
            nonlocal actual_timestamps
            actual_timestamps = ts

        listener = MagicMock()
        listener.on_response = on_response
        client.add_latency_listener(listener)
        price = {}
        timestamps = None

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'getSymbolPrice' and data['accountId'] == 'accountId' and data['symbol'] == 'AUDNZD' \
                    and data['application'] == 'RPC' and 'clientProcessingStarted' in data['timestamps']:
                nonlocal timestamps
                timestamps = deepcopy(data['timestamps'])
                timestamps['serverProcessingStarted'] = format_date(datetime.now())
                timestamps['serverProcessingFinished'] = format_date(datetime.now())
                timestamps['clientProcessingStarted'] = format_date(date(timestamps['clientProcessingStarted']))
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'price': price, 'timestamps': timestamps})

        await client.get_symbol_price('accountId', 'AUDNZD')
        await asyncio.sleep(0.05)
        assert account_id == 'accountId'
        assert request_type == 'getSymbolPrice'
        assert actual_timestamps['clientProcessingStarted'] == date(timestamps['clientProcessingStarted'])
        assert actual_timestamps['serverProcessingStarted'] == date(timestamps['serverProcessingStarted'])
        assert actual_timestamps['serverProcessingFinished'] == date(timestamps['serverProcessingFinished'])
        assert 'clientProcessingFinished' in actual_timestamps

    @pytest.mark.asyncio
    async def test_measure_price_latencies(self):
        """Should measure price streaming latencies."""
        prices = [{
            'symbol': 'AUDNZD',
            'timestamps': {
                'eventGenerated': format_date(datetime.now()),
                'serverProcessingStarted': format_date(datetime.now()),
                'serverProcessingFinished': format_date(datetime.now())
            }
        }]
        account_id = None
        symbol = None
        actual_timestamps = None
        listener = MagicMock()

        async def on_symbol_price(aid, sym, ts):
            nonlocal account_id
            account_id = aid
            nonlocal symbol
            symbol = sym
            nonlocal actual_timestamps
            actual_timestamps = ts
            await client.close()

        listener.on_symbol_price = on_symbol_price
        client.add_latency_listener(listener)
        await sio.emit('synchronization', {'type': 'prices', 'accountId': 'accountId', 'prices': prices,
                                           'equity': 100, 'margin': 200, 'freeMargin': 400, 'marginLevel': 40000})
        await client._socket.wait()
        assert account_id == 'accountId'
        assert symbol == 'AUDNZD'
        assert actual_timestamps['serverProcessingFinished'] == \
            date(prices[0]['timestamps']['serverProcessingFinished'])
        assert actual_timestamps['serverProcessingStarted'] == \
               date(prices[0]['timestamps']['serverProcessingStarted'])
        assert actual_timestamps['eventGenerated'] == \
               date(prices[0]['timestamps']['eventGenerated'])
        assert 'clientProcessingFinished' in actual_timestamps

    @pytest.mark.asyncio
    async def test_measure_update_latencies(self):
        """Should measure update latencies."""
        update = {
            'timestamps': {
                'eventGenerated': format_date(datetime.now()),
                'serverProcessingStarted': format_date(datetime.now()),
                'serverProcessingFinished': format_date(datetime.now())
            }
        }
        account_id = None
        actual_timestamps = None
        listener = MagicMock()

        async def on_update(aid, ts):
            nonlocal account_id
            account_id = aid
            nonlocal actual_timestamps
            actual_timestamps = ts
            await client.close()

        listener.on_update = on_update
        client.add_latency_listener(listener)
        await sio.emit('synchronization', {'type': 'update', 'accountId': 'accountId', **update})
        await client._socket.wait()
        assert account_id == 'accountId'
        assert actual_timestamps['serverProcessingFinished'] == \
               date(update['timestamps']['serverProcessingFinished'])
        assert actual_timestamps['serverProcessingStarted'] == \
               date(update['timestamps']['serverProcessingStarted'])
        assert actual_timestamps['eventGenerated'] == \
               date(update['timestamps']['eventGenerated'])
        assert 'clientProcessingFinished' in actual_timestamps

    @pytest.mark.asyncio
    async def test_process_trade_latency(self):
        """Should proces trade latency."""
        trade = {}
        response = {
            'numericCode': 10009,
            'stringCode': 'TRADE_RETCODE_DONE',
            'message': 'Request completed',
            'orderId': '46870472'
        }
        timestamps = {
            'clientExecutionStarted': format_date(datetime.now()),
            'serverExecutionStarted': format_date(datetime.now()),
            'serverExecutionFinished': format_date(datetime.now()),
            'tradeExecuted': format_date(datetime.now())
        }
        account_id = None
        actual_timestamps = None
        listener = MagicMock()

        def on_trade(aid, ts):
            nonlocal account_id
            account_id = aid
            nonlocal actual_timestamps
            actual_timestamps = ts

        listener.on_trade = on_trade
        client.add_latency_listener(listener)

        @sio.on('request')
        async def on_request(sid, data):
            assert data['trade'] == trade
            if data['type'] == 'trade' and data['accountId'] == 'accountId' and data['application'] == 'application':
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'response': response,
                                            'timestamps': timestamps})

        await client.trade('accountId', trade)
        assert account_id == 'accountId'
        assert actual_timestamps['clientExecutionStarted'] == \
               date(timestamps['clientExecutionStarted'])
        assert actual_timestamps['serverExecutionStarted'] == \
               date(timestamps['serverExecutionStarted'])
        assert actual_timestamps['serverExecutionFinished'] == \
               date(timestamps['serverExecutionFinished'])
        assert actual_timestamps['tradeExecuted'] == \
               date(timestamps['tradeExecuted'])
        assert 'clientProcessingFinished' in actual_timestamps

    @pytest.mark.asyncio
    async def test_reconnect(self):
        """Should reconnect to server on disconnect."""

        trade = {
            'actionType': 'ORDER_TYPE_SELL',
            'symbol': 'AUDNZD',
            'volume': 0.07
        }
        response = {
            'numericCode': 10009,
            'stringCode': 'TRADE_RETCODE_DONE',
            'message': 'Request completed',
            'orderId': '46870472'
        }
        listener = MagicMock()
        listener.on_reconnected = MagicMock()
        client.add_reconnect_listener(listener)
        request_counter = 0

        @sio.on('request')
        async def on_request(sid, data):
            print(data)
            if data['type'] == 'trade':
                nonlocal request_counter
                request_counter += 1
                await sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                            'requestId': data['requestId'], 'response': response})
            await sio.disconnect(sid)

        await client.trade('accountId', trade)
        await asyncio.sleep(0.1)
        listener.on_reconnected.assert_called_once()
        await client.trade('accountId', trade)
        assert request_counter == 2
        await client.close()



from ..clients.httpClient import HttpClient
from ..clients.copyFactory.configuration_client import ConfigurationClient
from ..clients.copyFactory.history_client import HistoryClient
from ..clients.copyFactory.trading_client import TradingClient
from ..metaApi.metaApi import RetryOpts
from typing_extensions import TypedDict
from typing import Optional


class CopyFactoryOpts(TypedDict):
    """CopyFactory options"""
    domain: str
    """Domain to connect to."""
    requestTimeout: float
    """Timeout for http requests in seconds."""
    retryOpts: Optional[RetryOpts]
    """Options for request retries."""


class CopyFactory:
    """MetaApi CopyFactory copy trading API SDK"""

    def __init__(self, token: str, opts: CopyFactoryOpts = None):
        """Inits CopyFactory class instance.

        Args:
            token: Authorization token.
            opts: Connection options.
        """
        opts: CopyFactoryOpts = opts or {}
        domain = opts['domain'] if 'domain' in opts else 'agiliumtrade.agiliumtrade.ai'
        request_timeout = opts['requestTimeout'] if 'requestTimeout' in opts else 60
        retry_opts = opts['retryOpts'] if 'retryOpts' in opts else {}
        http_client = HttpClient(request_timeout, retry_opts)
        self._configurationClient = ConfigurationClient(http_client, token, domain)
        self._historyClient = HistoryClient(http_client, token, domain)
        self._tradingClient = TradingClient(http_client, token, domain)

    @property
    def configuration_api(self) -> ConfigurationClient:
        """Returns CopyFactory configuration API.

        Returns:
            Configuration API.
        """
        return self._configurationClient

    @property
    def history_api(self) -> HistoryClient:
        """Returns CopyFactory history API.

        Returns:
            History API.
        """
        return self._historyClient

    @property
    def trading_api(self) -> TradingClient:
        """Returns CopyFactory history API.

        Returns:
            History API.
        """
        return self._tradingClient



import json
import mock as mock
import pytest
import respx
from httpx import Response
from ..httpClient import HttpClient
from .provisioningProfile_client import ProvisioningProfileClient

PROVISIONING_API_URL = 'https://mt-provisioning-api-v1.agiliumtrade.agiliumtrade.ai'
httpClient = HttpClient()
provisioning_client = ProvisioningProfileClient(httpClient, 'header.payload.sign')


class TestProvisioningProfileClient:
    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_many(self):
        """Should retrieve provisioning profiles from API."""
        expected = [{
            '_id': 'id',
            'name': 'name',
            'version': 4,
            'status': 'active'
        }]
        rsps = respx.get(f'{PROVISIONING_API_URL}/users/current/provisioning-profiles') \
            .mock(return_value=Response(200, json=expected))
        profiles = await provisioning_client.get_provisioning_profiles(5, 'active')
        assert rsps.calls[0].request.url == \
            f'{PROVISIONING_API_URL}/users/current/provisioning-profiles?version=5&status=active'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert profiles == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_profiles_with_account_token(self):
        """Should not retrieve provisioning profiles from API with account token."""
        provisioning_client = ProvisioningProfileClient(httpClient, 'token')
        try:
            await provisioning_client.get_provisioning_profiles(5, 'active')
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_provisioning_profiles method, because you ' + \
                'have connected with account access token. Please use API access token from ' + \
                'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_one(self):
        """Should retrieve provisioning profile from API."""
        expected = {
            '_id': 'id',
            'name': 'name',
            'version': 4,
            'status': 'active'
        }
        rsps = respx.get(f'{PROVISIONING_API_URL}/users/current/provisioning-profiles/id') \
            .mock(return_value=Response(200, json=expected))
        profile = await provisioning_client.get_provisioning_profile('id')
        assert rsps.calls[0].request.url == \
            f'{PROVISIONING_API_URL}/users/current/provisioning-profiles/id'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert profile == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_profile_with_account_token(self):
        """Should not retrieve provisioning profile from API with account token."""
        provisioning_client = ProvisioningProfileClient(httpClient, 'token')
        try:
            await provisioning_client.get_provisioning_profile('id')
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_provisioning_profile method, because you ' + \
                   'have connected with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_create(self):
        """Should create provisioning profile via API."""
        expected = {
            'id': 'id'
        }
        profile = {
            'name': 'name',
            'version': 4,
        }
        rsps = respx.post(f'{PROVISIONING_API_URL}/users/current/provisioning-profiles') \
            .mock(return_value=Response(200, json=expected))
        id = await provisioning_client.create_provisioning_profile(profile)
        assert rsps.calls[0].request.url == f'{PROVISIONING_API_URL}/users/current/provisioning-profiles'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert rsps.calls[0].request.content == json.dumps(profile).encode('utf-8')
        assert id == expected

    @pytest.mark.asyncio
    async def test_not_create_profile_with_account_token(self):
        """Should not create provisioning profile from API with account token."""
        provisioning_client = ProvisioningProfileClient(httpClient, 'token')
        try:
            await provisioning_client.create_provisioning_profile({})
        except Exception as err:
            assert err.__str__() == 'You can not invoke create_provisioning_profile method, because you ' + \
                   'have connected with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_upload_file(self):
        """Should upload file to a provisioning profile via API."""
        rsps = respx.put(f'{PROVISIONING_API_URL}/users/current/provisioning-profiles/id/servers.dat') \
            .mock(return_value=Response(204))
        with mock.patch('__main__.open', new=mock.mock_open(read_data='test')) as file:
            file.return_value = 'test', 'test2'
            await provisioning_client.upload_provisioning_profile_file('id', 'servers.dat', file())
            assert rsps.calls[0].request.url == \
                f'{PROVISIONING_API_URL}/users/current/provisioning-profiles/id/servers.dat'
            assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'

    @pytest.mark.asyncio
    async def test_not_upload_file_with_account_token(self):
        """Should not upload provisioning profile file via API with account token."""
        provisioning_client = ProvisioningProfileClient(httpClient, 'token')
        try:
            await provisioning_client.upload_provisioning_profile_file('id', 'servers.dat', {})
        except Exception as err:
            assert err.__str__() == 'You can not invoke upload_provisioning_profile_file method, because you ' + \
                   'have connected with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_delete(self):
        """Should delete provisioning profile via API."""
        rsps = respx.delete(f'{PROVISIONING_API_URL}/users/current/provisioning-profiles/id') \
            .mock(return_value=Response(200))
        await provisioning_client.delete_provisioning_profile('id')
        assert rsps.calls[0].request.url == \
            f'{PROVISIONING_API_URL}/users/current/provisioning-profiles/id'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'

    @pytest.mark.asyncio
    async def test_not_delete_with_account_token(self):
        """Should not delete provisioning profile via API with account token."""
        provisioning_client = ProvisioningProfileClient(httpClient, 'token')
        try:
            await provisioning_client.delete_provisioning_profile('id')
        except Exception as err:
            assert err.__str__() == 'You can not invoke delete_provisioning_profile method, because you ' + \
                   'have connected with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_update(self):
        """Should update provisioning profile via API."""
        rsps = respx.put(f'{PROVISIONING_API_URL}/users/current/provisioning-profiles/id') \
            .mock(return_value=Response(200))
        await provisioning_client.update_provisioning_profile('id', {'name': 'new name'})
        assert rsps.calls[0].request.url == \
               f'{PROVISIONING_API_URL}/users/current/provisioning-profiles/id'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert rsps.calls[0].request.content == json.dumps({'name': 'new name'}).encode('utf-8')

    @pytest.mark.asyncio
    async def test_not_update_with_account_token(self):
        """Should not update provisioning profile via API with account token."""
        provisioning_client = ProvisioningProfileClient(httpClient, 'token')
        try:
            await provisioning_client.update_provisioning_profile('id', {'name': 'new name'})
        except Exception as err:
            assert err.__str__() == 'You can not invoke update_provisioning_profile method, because you ' + \
                   'have connected with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'



from .provisioningProfile import ProvisioningProfile
from ..clients.metaApi.provisioningProfile_client import ProvisioningProfileClient, NewProvisioningProfileDto
from typing import List
from copy import deepcopy


class ProvisioningProfileApi:
    """Exposes provisioning profile API logic to the consumers."""

    def __init__(self, provisioning_profile_client: ProvisioningProfileClient):
        """Inits a provisioning profile API instance.

        Args:
            provisioning_profile_client: Provisioning profile REST API client.
        """
        self._provisioningProfileClient = provisioning_profile_client

    async def get_provisioning_profiles(self, version: int = None, status: str = None) -> List[ProvisioningProfile]:
        """Retrieves provisioning profiles

        Args:
            version: Optional version filter (allowed values are 4 and 5).
            status: Optional status filter (allowed values are new and active).

        Returns:
            A coroutine resolving with an array of provisioning profile entities.
        """

        profiles = await self._provisioningProfileClient.get_provisioning_profiles(version, status)
        return list(map(lambda profile: ProvisioningProfile(profile, self._provisioningProfileClient), profiles))

    async def get_provisioning_profile(self, provisioning_profile_id: str) -> ProvisioningProfile:
        """Retrieves a provisioning profile by id.

        Args:
            provisioning_profile_id: Provisioning profile id.

        Returns:
            A coroutine resolving with provisioning profile entity.
        """
        profile = await self._provisioningProfileClient.get_provisioning_profile(provisioning_profile_id)
        return ProvisioningProfile(profile, self._provisioningProfileClient)

    async def create_provisioning_profile(self, profile: NewProvisioningProfileDto) -> ProvisioningProfile:
        """Creates a provisioning profile.

        Args:
            profile: Provisioning profile data.

        Returns:
            A coroutine resolving with provisioning profile entity.
        """

        id = await self._provisioningProfileClient.create_provisioning_profile(profile)
        new_profile = deepcopy(profile)
        new_profile['_id'] = id['id']
        new_profile['status'] = 'new'
        return ProvisioningProfile(new_profile, self._provisioningProfileClient)



from .errorHandler import UnauthorizedException, ForbiddenException, ApiException, ConflictException, \
    ValidationException, InternalException, NotFoundException, TooManyRequestsException
from typing_extensions import TypedDict
from typing import Optional
from ..metaApi.models import ExceptionMessage
import json
import asyncio
import sys
import httpx
from httpx import HTTPError, Response


if sys.version_info[0] == 3 and sys.version_info[1] >= 8 and sys.platform.startswith('win'):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


class RequestOptions(TypedDict):
    """Options for HttpClient requests."""
    method: Optional[str]
    url: str
    headers: Optional[dict]
    params: Optional[dict]
    body: Optional[dict]
    files: Optional[dict]


class HttpClient:
    """HTTP client library based on requests module."""
    def __init__(self, timeout: float = 60, retry_opts=None):
        """Inits HttpClient class instance.

        Args:
            timeout: Request timeout in seconds.
        """
        if retry_opts is None:
            retry_opts = {}
        self._timeout = timeout
        self._retries = retry_opts['retries'] if 'retries' in retry_opts else 5
        self._minRetryDelayInSeconds = retry_opts['minDelayInSeconds'] if 'minDelayInSeconds' in retry_opts else 1
        self._maxRetryDelayInSeconds = retry_opts['maxDelayInSeconds'] if 'maxDelayInSeconds' in retry_opts else 30

    async def request(self, options: RequestOptions, retry_counter: int = 0) -> Response:
        """Performs a request. Response errors are returned as ApiError or subclasses.

        Args:
            options: Request options.

        Returns:
            A request response.
        """
        try:
            response = await self._make_request(options)
            response.raise_for_status()
            if response.content:
                try:
                    response = response.json()
                except Exception as err:
                    print('Error parsing json', err)
        except HTTPError as err:
            if err.__class__.__name__ == 'ConnectTimeout':
                error = err
            else:
                error = self._convert_error(err)
            if error.__class__.__name__ in ['ConflictException', 'InternalException', 'ApiException', 'ConnectTimeout']\
                    and retry_counter < self._retries:
                await asyncio.sleep(min(pow(2, retry_counter) * self._minRetryDelayInSeconds,
                                        self._maxRetryDelayInSeconds))
                return await self.request(options, retry_counter + 1)
            else:
                raise error
        return response

    async def _make_request(self, options: RequestOptions) -> Response:
        async with httpx.AsyncClient(timeout=self._timeout) as client:
            method = options['method'] if ('method' in options) else 'GET'
            url = options['url']
            params = options['params'] if 'params' in options else None
            files = options['files'] if 'files' in options else None
            headers = options['headers'] if 'headers' in options else None
            body = options['body'] if 'body' in options else None
            req = client.build_request(method, url, params=params, files=files, headers=headers, json=body)
            response = await client.send(req)
            return response

    def _convert_error(self, err: HTTPError):
        try:
            response: ExceptionMessage or TypedDict = json.loads(err.response.text)
        except Exception:
            response = {}
        err_message = response['message'] if 'message' in response else err.response.reason_phrase
        status = err.response.status_code
        if status == 400:
            details = response['details'] if 'details' in response else []
            return ValidationException(err_message, details)
        elif status == 401:
            return UnauthorizedException(err_message)
        elif status == 403:
            return ForbiddenException(err_message)
        elif status == 404:
            return NotFoundException(err_message)
        elif status == 409:
            return ConflictException(err_message)
        elif status == 429:
            err_metadata = response['metadata'] if 'metadata' in response else {}
            return TooManyRequestsException(err_message, err_metadata)
        elif status == 500:
            return InternalException(err_message)
        else:
            return ApiException(err_message, status)



from .. import MetaApi
import os
import pytest
import asyncio

token = os.getenv('TOKEN')
login = os.getenv('LOGIN_MT4')
password = os.getenv('PASSWORD_MT4')
server_name = os.getenv('SERVER_MT4') or 'Tradeview-Demo'
broker_srv_file = os.getenv('PATH_TO_BROKER_SRV') or './lib/integration_tests/files/tradeview-demo.broker.srv'
api: MetaApi = None


async def check_positions(connection):
    return {'local': len(connection.terminal_state.positions), 'real': len(await connection.get_positions())}


@pytest.fixture(autouse=True)
async def run_around_tests():
    global api
    api = MetaApi(token, {'application': 'MetaApi', 'domain': 'project-stock.v2.agiliumlabs.cloud'})
    yield


class TestSyncPositions:
    @pytest.mark.asyncio
    async def test_show_correct_positions(self):
        """Should show correct positions amount after opening and closing."""
        if token and login:
            profiles = await api.provisioning_profile_api.get_provisioning_profiles()
            profile = None
            for item in profiles:
                if item.name == server_name:
                    profile = item
                    break
            if not profile:
                profile = await api.provisioning_profile_api.create_provisioning_profile({
                    'name': server_name,
                    'version': 4,
                    'brokerTimezone': 'EET',
                    'brokerDSTSwitchTimezone': 'EET'
                })
                await profile.upload_file('broker.srv', broker_srv_file)
            if profile and profile.status == 'new':
                await profile.upload_file('broker.srv', broker_srv_file)
            accounts = await api.metatrader_account_api.get_accounts()
            account = None
            for item in accounts:
                if item.login == login and item.type.startswith('cloud-g2'):
                    account = item
                    break
            if not account:
                account = await api.metatrader_account_api.create_account({
                    'name': 'Test account-mt4',
                    'type': 'cloud-g2',
                    'login': login,
                    'password': password,
                    'server': server_name,
                    'provisioningProfileId': profile.id,
                    'application': 'MetaApi',
                    'magic': 1000
                })
            await account.deploy()
            await account.wait_connected()
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 600})
            start_positions = len(connection.terminal_state.positions)
            position_ids = []
            positions = await check_positions(connection)
            assert positions['local'] == positions['real']
            for i in range(10):
                result = await connection.create_market_buy_order('GBPUSD', 0.01, 0.9, 2.0)
                position_ids.append(result['positionId'])
                await asyncio.sleep(0.2)
            positions = await check_positions(connection)
            await asyncio.sleep(3)
            assert positions['local'] == start_positions + 10
            assert positions['real'] == start_positions + 10
            await asyncio.sleep(5)
            await asyncio.gather(*list(map(lambda id: connection.close_position(id), position_ids)))
            await asyncio.sleep(1)

            async def close_position_test(id):
                try:
                    await connection.get_position(id)
                    pytest.fail()
                except Exception as err:
                    pass
            await asyncio.gather(*list(map(lambda id: close_position_test(id), position_ids)))
            positions = await check_positions(connection)
            await asyncio.sleep(3)
            assert positions['local'] == start_positions
            assert positions['real'] == start_positions
            await account.undeploy()



from .metaApiConnection import MetaApiConnection
from ..clients.metaApi.metaApiWebsocket_client import MetaApiWebsocketClient
from .metatraderAccountModel import MetatraderAccountModel
from .historyStorage import HistoryStorage
from .connectionRegistryModel import ConnectionRegistryModel
from datetime import datetime
import asyncio


class ConnectionRegistry(ConnectionRegistryModel):
    """Manages account connections"""

    def __init__(self, meta_api_websocket_client: MetaApiWebsocketClient, application: str = 'MetaApi'):
        """Inits a MetaTrader connection registry instance.

        Args:
            meta_api_websocket_client: MetaApi websocket client.
            application: Application type.
        """
        self._meta_api_websocket_client = meta_api_websocket_client
        self._application = application
        self._connections = {}
        self._connectionLocks = {}

    async def connect(self, account: MetatraderAccountModel, history_storage: HistoryStorage,
                      history_start_time: datetime = None) -> MetaApiConnection:
        """Creates and returns a new account connection if doesnt exist, otherwise returns old.

        Args:
            account: MetaTrader account to connect to.
            history_storage: Terminal history storage.
            history_start_time: History start time.

        Returns:
            A coroutine resolving with account connection.
        """
        if account.id in self._connections:
            return self._connections[account.id]
        else:
            while account.id in self._connectionLocks:
                await self._connectionLocks[account.id]['promise']
            if account.id in self._connections:
                return self._connections[account.id]
            connection_lock = asyncio.Future()
            self._connectionLocks[account.id] = {'promise': connection_lock}
            connection = MetaApiConnection(self._meta_api_websocket_client, account, history_storage, self,
                                           history_start_time)
            try:
                await connection.initialize()
                await connection.subscribe()
                self._connections[account.id] = connection
            finally:
                del self._connectionLocks[account.id]
                connection_lock.set_result(None)
            return connection

    def remove(self, account_id: str):
        """Removes an account from registry.

        Args:
            account_id: MetaTrader account id to remove.
        """
        if account_id in self._connections:
            del self._connections[account_id]

    @property
    def application(self) -> str:
        """Returns application type.

        Returns:
            Application type.
        """
        return self._application



class NotSynchronizedException(Exception):
    """Exception which indicates that MetaApi MetaTrader account was not synchronized yet. See
    https://metaapi.cloud/docs/client/websocket/synchronizationMode/ for more details"""

    def __init__(self, message: str):
        """Inits the exception.

        Args:
            message: Exception message.
        """
        super().__init__(message)



from ..clients.metaApi.latencyListener import LatencyListener, ResponseTimestamps, SymbolPriceTimestamps, \
    UpdateTimestamps, TradeTimestamps
from .reservoir.statisticalReservoir import StatisticalReservoir
from .reservoir.reservoir import Reservoir
from typing import Dict


class LatencyMonitor(LatencyListener):
    """Responsible for monitoring MetaApi application latencies."""

    def __init__(self):
        """Inits latency monitor instance."""
        super().__init__()
        self._tradeReservoirs = {
            'clientLatency': self._initialize_reservoirs(),
            'serverLatency': self._initialize_reservoirs(),
            'brokerLatency': self._initialize_reservoirs()
        }
        self._updateReservoirs = {
            'clientLatency': self._initialize_reservoirs(),
            'serverLatency': self._initialize_reservoirs(),
            'brokerLatency': self._initialize_reservoirs()
        }
        self._priceReservoirs = {
            'clientLatency': self._initialize_reservoirs(),
            'serverLatency': self._initialize_reservoirs(),
            'brokerLatency': self._initialize_reservoirs()
        }
        self._requestReservoirs = {
            'branch': True
        }

    async def on_response(self, account_id: str, type: str, timestamps: ResponseTimestamps):
        """Invoked with latency information when application receives a response to RPC request.

        Args:
            account_id: Account id.
            type: Request type.
            timestamps: Request timestamps object containing latency information.

        Returns:
            A coroutine which resolves when latency information is processed."""
        if type not in self._requestReservoirs:
            self._requestReservoirs[type] = {
                'branch': True,
                'clientLatency': self._initialize_reservoirs(),
                'serverLatency': self._initialize_reservoirs()
            }
        if 'serverProcessingStarted' in timestamps and 'serverProcessingFinished' in timestamps:
            server_latency = timestamps['serverProcessingFinished'].timestamp() - \
                timestamps['serverProcessingStarted'].timestamp()
            self._save_measurement(self._requestReservoirs[type]['serverLatency'], server_latency)

        if 'clientProcessingStarted' in timestamps and 'clientProcessingFinished' in timestamps and \
                'serverProcessingStarted' in timestamps and 'serverProcessingFinished' in timestamps:
            server_latency = timestamps['serverProcessingFinished'].timestamp() - \
                timestamps['serverProcessingStarted'].timestamp()
            client_latency = timestamps['clientProcessingFinished'].timestamp() - \
                timestamps['clientProcessingStarted'].timestamp() - server_latency
            self._save_measurement(self._requestReservoirs[type]['clientLatency'], client_latency)

    @property
    def request_latencies(self) -> Dict:
        """Returns request processing latencies.

        Returns:
            Request processing latencies.
        """
        return self._construct_latencies_recursively(self._requestReservoirs)

    async def on_symbol_price(self, account_id: str, symbol: str, timestamps: SymbolPriceTimestamps):
        """Invoked with latency information when application receives symbol price update event.

        Args:
            account_id: Account id.
            symbol: Price symbol.
            timestamps: Timestamps object containing latency information about price streaming.

        Returns:
            A coroutine which resolves when latency information is processed.
        """
        if 'eventGenerated' in timestamps and 'serverProcessingStarted' in timestamps:
            broker_latency = timestamps['serverProcessingStarted'].timestamp() - \
                             timestamps['eventGenerated'].timestamp()
            self._save_measurement(self._priceReservoirs['brokerLatency'], broker_latency)
        if 'serverProcessingStarted' in timestamps and 'serverProcessingFinished' in timestamps:
            server_latency = timestamps['serverProcessingFinished'].timestamp() - \
                             timestamps['serverProcessingStarted'].timestamp()
            self._save_measurement(self._priceReservoirs['serverLatency'], server_latency)
        if 'serverProcessingFinished' in timestamps and 'clientProcessingFinished' in timestamps:
            client_latency = timestamps['clientProcessingFinished'].timestamp() - \
                             timestamps['serverProcessingFinished'].timestamp()
            self._save_measurement(self._priceReservoirs['clientLatency'], client_latency)

    @property
    def price_latencies(self) -> Dict:
        """Returns price streaming latencies.

        Returns:
            Price streaming latencies.
        """
        return self._construct_latencies_recursively(self._priceReservoirs)

    async def on_update(self, account_id: str, timestamps: UpdateTimestamps):
        """Invoked with latency information when application receives update event.

        Args:
            account_id: Account id.
            timestamps: Timestamps object containing latency information about update streaming.

        Returns:
            A coroutine which resolves when latency information is processed."""
        if 'eventGenerated' in timestamps and 'serverProcessingStarted' in timestamps:
            broker_latency = timestamps['serverProcessingStarted'].timestamp() - \
                             timestamps['eventGenerated'].timestamp()
            self._save_measurement(self._updateReservoirs['brokerLatency'], broker_latency)
        if 'serverProcessingStarted' in timestamps and 'serverProcessingFinished' in timestamps:
            server_latency = timestamps['serverProcessingFinished'].timestamp() - \
                             timestamps['serverProcessingStarted'].timestamp()
            self._save_measurement(self._updateReservoirs['serverLatency'], server_latency)
        if 'serverProcessingFinished' in timestamps and 'clientProcessingFinished' in timestamps:
            client_latency = timestamps['clientProcessingFinished'].timestamp() - \
                             timestamps['serverProcessingFinished'].timestamp()
            self._save_measurement(self._updateReservoirs['clientLatency'], client_latency)

    @property
    def update_latencies(self) -> Dict:
        """Returns update streaming latencies.

        Returns:
            Update streaming latencies.
        """
        return self._construct_latencies_recursively(self._updateReservoirs)

    async def on_trade(self, account_id: str, timestamps: TradeTimestamps):
        """Invoked with latency information when application receives trade response.

        Args:
            account_id: Account id.
            timestamps: Timestamps object containing latency information about a trade.

        Returns:
            A coroutine which resolves when latency information is processed."""
        if 'clientProcessingStarted' in timestamps and 'serverProcessingStarted' in timestamps:
            client_latency = timestamps['serverProcessingStarted'].timestamp() - \
                             timestamps['clientProcessingStarted'].timestamp()
            self._save_measurement(self._tradeReservoirs['clientLatency'], client_latency)
        if 'serverProcessingStarted' in timestamps and 'tradeStarted' in timestamps:
            server_latency = timestamps['tradeStarted'].timestamp() - \
                             timestamps['serverProcessingStarted'].timestamp()
            self._save_measurement(self._tradeReservoirs['serverLatency'], server_latency)
        if 'tradeStarted' in timestamps and 'tradeExecuted' in timestamps:
            broker_latency = timestamps['tradeExecuted'].timestamp() - \
                             timestamps['tradeStarted'].timestamp()
            self._save_measurement(self._tradeReservoirs['brokerLatency'], broker_latency)

    @property
    def trade_latencies(self) -> Dict:
        """Returns trade latencies.

        Returns:
            Trade latencies.
        """
        return self._construct_latencies_recursively(self._tradeReservoirs)

    def _save_measurement(self, reservoirs, client_latency: float):
        client_latency = int(client_latency * 1000)
        for key in reservoirs:
            e = reservoirs[key]
            if key == 'branch':
                continue
            e['percentiles'].push_measurement(client_latency)
            e['reservoir'].push_measurement(client_latency)

    def _construct_latencies_recursively(self, reservoirs):
        result = {}
        for key in reservoirs:
            e = reservoirs[key]
            if key == 'branch':
                continue
            result[key] = self._construct_latencies_recursively(e) if ('branch' in e and e['branch']) else {
                'p50': e['percentiles'].get_percentile(50),
                'p75': e['percentiles'].get_percentile(75),
                'p90': e['percentiles'].get_percentile(90),
                'p95': e['percentiles'].get_percentile(95),
                'p98': e['percentiles'].get_percentile(98),
                'avg': e['reservoir'].get_statistics()['average'],
                'count': e['reservoir'].get_statistics()['count'],
                'min': e['reservoir'].get_statistics()['min'],
                'max': e['reservoir'].get_statistics()['max']
            }
        return result

    def _initialize_reservoirs(self):
        return {
            'branch': True,
            '1h': {
                'percentiles': StatisticalReservoir(1000, 60 * 60 * 1000),
                'reservoir': Reservoir(60, 60 * 60 * 1000)
            },
            '1d': {
                'percentiles': StatisticalReservoir(1000, 24 * 60 * 60 * 1000),
                'reservoir': Reservoir(60, 24 * 60 * 60 * 1000)
            },
            '1w': {
                'percentiles': StatisticalReservoir(1000, 7 * 24 * 60 * 60 * 1000),
                'reservoir': Reservoir(60, 7 * 24 * 60 * 60 * 1000)
            },
        }



import os
import asyncio
from metaapi_cloud_sdk import MetaApi

from metaapi_cloud_sdk.clients.metaApi.tradeException import TradeException
# Note: for information on how to use this example code please read https://metaapi.cloud/docs/client/usingCodeExamples

token = os.getenv('TOKEN') or '<put in your token here>'
login = os.getenv('LOGIN') or '<put in your MT login here>'
password = os.getenv('PASSWORD') or '<put in your MT password here>'
server_name = os.getenv('SERVER') or '<put in your MT server name here>'
server_dat_file = os.getenv('PATH_TO_SERVERS_DAT') or '/path/to/your/servers.dat'


async def test_meta_api_synchronization():
    api = MetaApi(token)
    try:
        profiles = await api.provisioning_profile_api.get_provisioning_profiles()

        # create test MetaTrader account profile
        profile = None
        for item in profiles:
            if item.name == server_name:
                profile = item
                break
        if not profile:
            print('Creating account profile')
            profile = await api.provisioning_profile_api.create_provisioning_profile({
                'name': server_name,
                'version': 5,
                'brokerTimezone': 'EET',
                'brokerDSTSwitchTimezone': 'EET'
            })
            await profile.upload_file('servers.dat', server_dat_file)
        if profile and profile.status == 'new':
            print('Uploading servers.dat')
            await profile.upload_file('servers.dat', server_dat_file)
        else:
            print('Account profile already created')

        # Add test MetaTrader account
        accounts = await api.metatrader_account_api.get_accounts()
        account = None
        for item in accounts:
            if item.login == login and item.type.startswith('cloud'):
                account = item
                break
        if not account:
            print('Adding MT5 account to MetaApi')
            account = await api.metatrader_account_api.create_account({
                'name': 'Test account',
                'type': 'cloud',
                'login': login,
                'password': password,
                'server': server_name,
                'provisioningProfileId': profile.id,
                'application': 'MetaApi',
                'magic': 1000
            })
        else:
            print('MT5 account already added to MetaApi')

        #  wait until account is deployed and connected to broker
        print('Deploying account')
        await account.deploy()
        print('Waiting for API server to connect to broker (may take couple of minutes)')
        await account.wait_connected()

        # connect to MetaApi API
        connection = await account.connect()

        # wait until terminal state synchronized to the local state
        print('Waiting for SDK to synchronize to terminal state (may take some time depending on your history size)')
        await connection.wait_synchronized({'timeoutInSeconds': 600})

        # access local copy of terminal state
        print('Testing terminal state access')
        terminal_state = connection.terminal_state
        print('connected:', terminal_state.connected)
        print('connected to broker:', terminal_state.connected_to_broker)
        print('account information:', terminal_state.account_information)
        print('positions:', terminal_state.positions)
        print('orders:', terminal_state.orders)
        print('specifications:', terminal_state.specifications)
        print('EURUSD specification:', terminal_state.specification('EURUSD'))
        print('EURUSD price:', terminal_state.price('EURUSD'))

        # trade
        print('Submitting pending order')
        try:
            result = await connection.create_limit_buy_order('GBPUSD', 0.07, 1.0, 0.9, 2.0,
                                                             {'comment': 'comm', 'clientId': 'TE_GBPUSD_7hyINWqAlE'})
            print('Trade successful, result code is ' + result['stringCode'])
        except Exception as err:
            print('Trade failed with error:')
            print(api.format_error(err))

        # finally, undeploy account after the test
        print('Undeploying MT5 account so that it does not consume any unwanted resources')
        await account.undeploy()

    except Exception as err:
        print(api.format_error(err))

asyncio.run(test_meta_api_synchronization())



from ..httpClient import HttpClient
from .trading_client import TradingClient
import pytest
from ...metaApi.models import date
import respx
from httpx import Response
copy_factory_api_url = 'https://trading-api-v1.agiliumtrade.agiliumtrade.ai'
http_client = HttpClient()
trading_client = TradingClient(http_client, 'header.payload.sign')


@pytest.fixture(autouse=True)
async def run_around_tests():
    global http_client
    global trading_client
    http_client = HttpClient()
    trading_client = TradingClient(http_client, 'header.payload.sign')


class TestTradingClient:
    @respx.mock
    @pytest.mark.asyncio
    async def test_resynchronize_copyfactory_account(self):
        rsps = respx.post(f'{copy_factory_api_url}/users/current/accounts/' +
                          '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef/' +
                          'resynchronize?strategyId=ABCD').mock(return_value=Response(200))
        await trading_client.resynchronize('0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
                                           ['ABCD'])
        assert rsps.calls[0].request.method == 'POST'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'

    @pytest.mark.asyncio
    async def test_not_resynchronize_account_with_account_token(self):
        """Should not resynchronize account with account token."""
        trading_client = TradingClient(http_client, 'token')
        try:
            await trading_client.resynchronize('0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
                                               ['ABCD'])
        except Exception as err:
            assert err.__str__() == 'You can not invoke resynchronize method, ' + \
                   'because you have connected with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_stopouts(self):
        """Should retrieve stopouts."""
        expected = [{
          'reason': 'max-drawdown',
          'stoppedAt': '2020-08-08T07:57:30.328Z',
          'strategy': {
            'id': 'ABCD',
            'name': 'Strategy'
          },
          'reasonDescription': 'total strategy equity drawdown exceeded limit'
        }]
        rsps = respx.get(f'{copy_factory_api_url}/users/current/accounts/' +
                         '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef/' +
                         'stopouts').mock(return_value=Response(200, json=expected))
        stopouts = await trading_client.get_stopouts('0123456789abcdef0123456789abcdef0123456789abcdef' +
                                                     '0123456789abcdef')
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/accounts/' + \
            '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef/stopouts'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert stopouts == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_stopouts_with_account_token(self):
        """Should not retrieve stopouts from API with account token."""
        trading_client = TradingClient(http_client, 'token')
        try:
            await trading_client.get_stopouts('0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef')
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_stopouts method, ' + \
                   'because you have connected with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_reset_stopouts(self):
        rsps = respx.post(f'{copy_factory_api_url}/users/current/accounts/' +
                          '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef/' +
                          'strategies-subscribed/ABCD/stopouts/daily-equity/reset').mock(return_value=Response(200))
        await trading_client.reset_stopouts('0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
                                            'ABCD', 'daily-equity')
        assert rsps.calls[0].request.method == 'POST'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'

    @pytest.mark.asyncio
    async def test_not_reset_stopouts_with_account_token(self):
        """Should not reset stopout with account token."""
        trading_client = TradingClient(http_client, 'token')
        try:
            await trading_client.reset_stopouts('0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
                                                'ABCD', 'daily-equity')
        except Exception as err:
            assert err.__str__() == 'You can not invoke reset_stopouts method, ' + \
                   'because you have connected with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_copy_trading_log(self):
        """Should retrieve copy trading user log."""
        expected = [{
          'time': '2020-08-08T07:57:30.328Z',
          'level': 'INFO',
          'message': 'message'
        }]
        rsps = respx.get(f'{copy_factory_api_url}/users/current/accounts/' +
                         '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef/user-log' +
                         '?offset=10&limit=100&startTime=2020-08-01T00%3A00%3A00.000Z&endTime=2020-0' +
                         '8-10T00%3A00%3A00.000Z').mock(return_value=Response(200, json=expected))
        records = await trading_client.get_user_log('0123456789abcdef0123456789abcdef0123456789abcdef' +
                                                    '0123456789abcdef', date('2020-08-01T00:00:00.000Z'),
                                                    date('2020-08-10T00:00:00.000Z'), 10, 100)
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/accounts/' + \
               '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef/user-log' + \
               '?offset=10&limit=100&startTime=2020-08-01T00%3A00%3A00.000Z&endTime=2020-08-10T00%3A00%3A00.000Z'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        expected[0]['time'] = date(expected[0]['time'])
        assert records == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_copy_trading_log_with_account_token(self):
        """Should not retrieve copy trading user log from API with account token."""
        trading_client = TradingClient(http_client, 'token')
        try:
            await trading_client.get_user_log('0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef')
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_user_log method, ' + \
                   'because you have connected with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'



from abc import ABC, abstractmethod


class ReconnectListener(ABC):
    """Defines interface for a websocket reconnect listener class."""

    @abstractmethod
    async def on_reconnected(self):
        """Invoked when connection to MetaTrader terminal re-established.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass



from ..clients.httpClient import HttpClient
from ..clients.metaApi.metaApiWebsocket_client import MetaApiWebsocketClient
from ..metaApi.provisioningProfileApi import ProvisioningProfileApi
from ..clients.metaApi.provisioningProfile_client import ProvisioningProfileClient
from ..metaApi.metatraderAccountApi import MetatraderAccountApi
from ..clients.metaApi.metatraderAccount_client import MetatraderAccountClient
from ..clients.metaApi.packetLogger import PacketLoggerOpts
from ..clients.errorHandler import ValidationException
from ..metaApi.connectionRegistry import ConnectionRegistry
from .metatraderDemoAccountApi import MetatraderDemoAccountApi
from ..clients.metaApi.metatraderDemoAccount_client import MetatraderDemoAccountClient
from .latencyMonitor import LatencyMonitor
import re
import traceback
from typing import Optional
from typing_extensions import TypedDict


class RetryOpts(TypedDict):
    retries: Optional[int]
    """Maximum amount of request retries, default value is 5."""
    minDelayInSeconds: Optional[float]
    """Minimum delay in seconds until request retry, default value is 1."""
    maxDelayInSeconds: Optional[float]
    """Maximum delay in seconds until request retry, default value is 30."""


class MetaApiOpts(TypedDict):
    """MetaApi options"""
    application: Optional[str]
    """Application id."""
    domain: Optional[str]
    """Domain to connect to, default is agiliumtrade.agiliumtrade.ai."""
    requestTimeout: Optional[float]
    """Timeout for socket requests in seconds."""
    connectTimeout: Optional[float]
    """Timeout for connecting to server in seconds."""
    packetOrderingTimeout: Optional[float]
    """Packet ordering timeout in seconds."""
    packetLogger: Optional[PacketLoggerOpts]
    """Packet logger options."""
    enableLatencyMonitor: Optional[bool]
    """An option to enable latency tracking."""
    enableLatencyTracking: Optional[bool]
    """An option to enable latency tracking."""
    maxConcurrentSynchronizations: Optional[int]
    """Max concurrent synchronizations via websocket client."""
    retryOpts: Optional[RetryOpts]
    """Options for request retries."""


class MetaApi:
    """MetaApi MetaTrader API SDK"""

    def __init__(self, token: str, opts: MetaApiOpts = None):
        """Inits MetaApi class instance.

        Args:
            token: Authorization token.
            opts: Application options.
        """
        opts: MetaApiOpts = opts or {}
        application = opts['application'] if 'application' in opts else 'MetaApi'
        domain = opts['domain'] if 'domain' in opts else 'agiliumtrade.agiliumtrade.ai'
        request_timeout = opts['requestTimeout'] if 'requestTimeout' in opts else 60
        connect_timeout = opts['connectTimeout'] if 'connectTimeout' in opts else 60
        packet_ordering_timeout = opts['packetOrderingTimeout'] if 'packetOrderingTimeout' in opts else 60
        retry_opts = opts['retryOpts'] if 'retryOpts' in opts else {}
        packet_logger = opts['packetLogger'] if 'packetLogger' in opts else {}
        max_concurrent_synchronizations = opts['maxConcurrentSynchronizations'] if 'maxConcurrentSynchronizations' in \
                                                                                   opts else 5
        if not re.search(r"[a-zA-Z0-9_]+", application):
            raise ValidationException('Application name must be non-empty string consisting ' +
                                      'from letters, digits and _ only')
        http_client = HttpClient(request_timeout, retry_opts)
        self._metaApiWebsocketClient = MetaApiWebsocketClient(
            token, {'application': application, 'domain': domain, 'requestTimeout': request_timeout,
                    'connectTimeout': connect_timeout, 'packetLogger': packet_logger,
                    'packetOrderingTimeout': packet_ordering_timeout,
                    'maxConcurrentSynchronizations': max_concurrent_synchronizations,
                    'retryOpts': retry_opts})
        self._provisioningProfileApi = ProvisioningProfileApi(ProvisioningProfileClient(http_client, token, domain))
        self._connectionRegistry = ConnectionRegistry(self._metaApiWebsocketClient, application)
        self._metatraderAccountApi = MetatraderAccountApi(MetatraderAccountClient(http_client, token, domain),
                                                          self._metaApiWebsocketClient, self._connectionRegistry)
        self._metatraderDemoAccountApi = MetatraderDemoAccountApi(MetatraderDemoAccountClient(http_client, token,
                                                                                              domain))
        if ('enableLatencyTracking' in opts and opts['enableLatencyTracking']) or ('enableLatencyMonitor' in opts and
                                                                                   opts['enableLatencyMonitor']):
            self._latencyMonitor = LatencyMonitor()
            self._metaApiWebsocketClient.add_latency_listener(self._latencyMonitor)

    @property
    def provisioning_profile_api(self) -> ProvisioningProfileApi:
        """Returns provisioning profile API.

        Returns:
            Provisioning profile API.
        """
        return self._provisioningProfileApi

    @property
    def metatrader_account_api(self) -> MetatraderAccountApi:
        """Returns MetaTrader account API.

        Returns:
            MetaTrader account API.
        """
        return self._metatraderAccountApi

    @property
    def metatrader_demo_account_api(self) -> MetatraderDemoAccountApi:
        """Returns MetaTrader demo account API.

        Returns:
            MetaTrader demo account API.
        """
        return self._metatraderDemoAccountApi

    @property
    def latency_monitor(self) -> LatencyMonitor:
        """Returns MetaApi application latency monitor.

        Returns:
            Latency monitor.
        """
        return self._latencyMonitor

    def format_error(self, err: Exception):
        """Formats and outputs metaApi errors with additional information.

        Args:
            err: Exception to process.
        """
        error = {'name': err.__class__.__name__, 'message': err.args[0]}
        if hasattr(err, 'status_code'):
            error['status_code'] = err.status_code
        if err.__class__.__name__ == 'ValidationException':
            error['details'] = err.details
        if err.__class__.__name__ == 'TradeException':
            error['string_code'] = err.stringCode
        if err.__class__.__name__ == 'TooManyRequestsException':
            error['metadata'] = err.metadata
        error['trace'] = traceback.format_exc()
        return error

    def close(self):
        """Closes all clients and connections"""
        self._metaApiWebsocketClient.remove_latency_listener(self._latencyMonitor)
        self._metaApiWebsocketClient.close()



from .terminalState import TerminalState
import pytest
from mock import patch
from asyncio import sleep
from datetime import datetime
state = TerminalState()


@pytest.fixture(autouse=True)
def run_around_tests():
    global state
    state = TerminalState()
    yield


class TestTerminalState:

    @pytest.mark.asyncio
    async def test_return_connection_state(self):
        """Should return connection state."""
        assert not state.connected
        await state.on_connected(1, 1)
        assert state.connected
        await state.on_disconnected(1)
        assert not state.connected

    @pytest.mark.asyncio
    async def test_return_broker_connection_state(self):
        """Should return broker connection state."""
        assert not state.connected_to_broker
        await state.on_broker_connection_status_changed(1, True)
        assert state.connected_to_broker
        await state.on_broker_connection_status_changed(1, False)
        assert not state.connected_to_broker
        await state.on_broker_connection_status_changed(1, True)
        await state.on_disconnected(1)
        assert not state.connected_to_broker

    @pytest.mark.asyncio
    async def test_call_disconnect(self):
        """Should call an on_disconnect if there was no signal for a long time"""
        with patch('lib.metaApi.terminalState.asyncio.sleep', new=lambda x: sleep(x / 50)):
            await state.on_connected(1, 1)
            await state.on_broker_connection_status_changed(1, True)
            await sleep(0.2)
            await state.on_broker_connection_status_changed(1, True)
            await sleep(1.1)
            assert state.connected_to_broker
            assert state.connected
            await sleep(0.2)
            assert not state.connected_to_broker
            assert not state.connected

    @pytest.mark.asyncio
    async def test_return_account_information(self):
        """Should return account information."""
        assert not state.account_information
        await state.on_account_information_updated(1, {'balance': 1000})
        assert state.account_information == {'balance': 1000}

    @pytest.mark.asyncio
    async def test_return_positions(self):
        """Should return positions."""
        assert len(state.positions) == 0
        await state.on_position_updated(1, {'id': '1', 'profit': 10})
        await state.on_position_updated(1, {'id': '2'})
        await state.on_position_updated(1, {'id': '1', 'profit': 11})
        assert len(state.positions) == 2
        await state.on_position_removed(1, '2')
        assert len(state.positions) == 1
        assert state.positions == [{'id': '1', 'profit': 11}]

    @pytest.mark.asyncio
    async def test_return_orders(self):
        """Should return orders."""
        assert len(state.orders) == 0
        await state.on_order_updated(1, {'id': '1', 'openPrice': 10})
        await state.on_order_updated(1, {'id': '2'})
        await state.on_order_updated(1, {'id': '1', 'openPrice': 11})
        assert len(state.orders) == 2
        await state.on_order_completed(1, '2')
        assert len(state.orders) == 1
        assert state.orders == [{'id': '1', 'openPrice': 11}]

    @pytest.mark.asyncio
    async def test_return_specifications(self):
        """Should return specifications."""
        assert len(state.specifications) == 0
        await state.on_symbol_specification_updated(1, {'symbol': 'EURUSD', 'tickSize': 0.00001})
        await state.on_symbol_specification_updated(1, {'symbol': 'GBPUSD'})
        await state.on_symbol_specification_updated(1, {'symbol': 'EURUSD', 'tickSize': 0.0001})
        assert len(state.specifications) == 2
        assert state.specifications == [{'symbol': 'EURUSD', 'tickSize': 0.0001}, {'symbol': 'GBPUSD'}]
        assert state.specification('EURUSD') == {'symbol': 'EURUSD', 'tickSize': 0.0001}

    @pytest.mark.asyncio
    async def test_return_price(self):
        """Should return price."""
        assert not state.price('EURUSD')
        await state.on_symbol_prices_updated(1, [{'time': datetime.fromtimestamp(1000000),
                                                  'symbol': 'EURUSD', 'bid': 1, 'ask': 1.1}])
        await state.on_symbol_prices_updated(1, [{'time': datetime.fromtimestamp(1000000), 'symbol': 'GBPUSD'}])
        await state.on_symbol_prices_updated(1, [{'time': datetime.fromtimestamp(1000000),
                                                  'symbol': 'EURUSD', 'bid': 1, 'ask': 1.2}])
        assert state.price('EURUSD') == {'time': datetime.fromtimestamp(1000000),
                                         'symbol': 'EURUSD', 'bid': 1, 'ask': 1.2}

    @pytest.mark.asyncio
    async def test_update_account_equity_and_position(self):
        """Should update account equity and position profit on price update."""
        await state.on_account_information_updated(1, {'equity': 1000, 'balance': 800})
        await state.on_positions_replaced(1, [{
            'id': '1',
            'symbol': 'EURUSD',
            'type': 'POSITION_TYPE_BUY',
            'currentPrice': 9,
            'currentTickValue': 0.5,
            'openPrice': 8,
            'profit': 100,
            'volume': 2
        }])
        await state.on_position_updated(1, {
            'id': '2',
            'symbol': 'AUDUSD',
            'type': 'POSITION_TYPE_BUY',
            'currentPrice': 9,
            'currentTickValue': 0.5,
            'openPrice': 8,
            'profit': 100,
            'volume': 2
        })
        await state.on_symbol_specification_updated(1, {'symbol': 'EURUSD', 'tickSize': 0.01})
        await state.on_symbol_specification_updated(1, {'symbol': 'AUDUSD', 'tickSize': 0.01})
        await state.on_symbol_prices_updated(1, [
          {
            'time': datetime.now(),
            'symbol': 'EURUSD',
            'profitTickValue': 0.5,
            'lossTickValue': 0.5,
            'bid': 10,
            'ask': 11
          },
          {
            'time': datetime.now(),
            'symbol': 'AUDUSD',
            'profitTickValue': 0.5,
            'lossTickValue': 0.5,
            'bid': 10,
            'ask': 11
          }
        ])
        assert list(map(lambda p: p['profit'], state.positions)) == [200, 200]
        assert list(map(lambda p: p['unrealizedProfit'], state.positions)) == [200, 200]
        assert list(map(lambda p: p['currentPrice'], state.positions)) == [10, 10]
        assert state.account_information['equity'] == 1200

    @pytest.mark.asyncio
    async def test_update_margin_fields(self):
        """Should update margin fields on price update."""
        await state.on_account_information_updated(1, {'equity': 1000, 'balance': 800})
        await state.on_symbol_prices_updated(1, [], 100, 200, 400, 40000)
        assert state.account_information['equity'] == 100
        assert state.account_information['margin'] == 200
        assert state.account_information['freeMargin'] == 400
        assert state.account_information['marginLevel'] == 40000

    @pytest.mark.asyncio
    async def test_update_order_current_price_on_price_update(self):
        """Should update order currentPrice on price update."""

        await state.on_order_updated(1, {
          'id': '1',
          'symbol': 'EURUSD',
          'type': 'ORDER_TYPE_BUY_LIMIT',
          'currentPrice': 9
        })
        await state.on_order_updated(1, {
            'id': '2',
            'symbol': 'AUDUSD',
            'type': 'ORDER_TYPE_SELL_LIMIT',
            'currentPrice': 9
        })
        await state.on_symbol_specification_updated(1, {'symbol': 'EURUSD', 'tickSize': 0.01})
        await state.on_symbol_prices_updated(1, [{
          'time': datetime.now(),
          'symbol': 'EURUSD',
          'profitTickValue': 0.5,
          'lossTickValue': 0.5,
          'bid': 10,
          'ask': 11
        }])
        assert list(map(lambda o: o['currentPrice'], state.orders)) == [11, 9]



from .provisioningProfile import ProvisioningProfile
from .provisioningProfileApi import ProvisioningProfileApi
from ..clients.metaApi.provisioningProfile_client import ProvisioningProfileClient, ProvisioningProfileDto
import pytest
from mock import MagicMock, AsyncMock


class MockClient(ProvisioningProfileClient):
    async def get_provisioning_profiles(self, version: int, status: str):
        pass

    async def get_provisioning_profile(self, id: str):
        pass

    async def create_provisioning_profile(self, provisioning_profile: ProvisioningProfileDto):
        pass

    async def delete_provisioning_profile(self, id: str):
        pass

    async def upload_provisioning_profile_file(self, provisioning_profile_id: str,
                                               file_name: str, file: str or memoryview):
        pass


client = MockClient(MagicMock(), 'token')
api = ProvisioningProfileApi(client)


class TestProvisioningProfileApi:

    @pytest.mark.asyncio
    async def test_retrieve_provisioning_profiles(self):
        """Should retrieve provisioning profiles."""

        client.get_provisioning_profiles = AsyncMock(return_value=[{'_id': 'id'}])
        profiles = await api.get_provisioning_profiles(4, 'new')
        for p in profiles:
            assert p.id == 'id'
            assert isinstance(p, ProvisioningProfile)

        client.get_provisioning_profiles.assert_called_with(4, 'new')

    @pytest.mark.asyncio
    async def test_retrieve_provisioning_profile(self):
        """Should retrieve provisioning profile by id."""

        client.get_provisioning_profile = AsyncMock(return_value={'_id': 'id', 'name': 'name',
                                                                  'version': 4, 'status': 'new',
                                                                  'brokerTimezone': 'EET',
                                                                  'brokerDSTSwitchTimezone': 'EET'})
        profile = await api.get_provisioning_profile('id')
        assert profile.id == 'id'
        assert profile.name == 'name'
        assert profile.version == 4
        assert profile.status == 'new'
        assert profile.broker_timezone == 'EET'
        assert profile.broker_dst_switch_timezone == 'EET'
        assert isinstance(profile, ProvisioningProfile)
        client.get_provisioning_profile.assert_called_with('id')

    @pytest.mark.asyncio
    async def test_create_provisioning_profile(self):
        """Should create provisioning profile."""

        client.create_provisioning_profile = AsyncMock(return_value={'id': 'id'})
        profile = await api.create_provisioning_profile({'name': 'name', 'version': 4, 'brokerTimezone': 'EET',
                                                         'brokerDSTSwitchTimezone': 'EET'})
        assert profile.id == 'id'
        assert profile.name == 'name'
        assert profile.version == 4
        assert profile.status == 'new'
        assert profile.broker_timezone == 'EET'
        assert profile.broker_dst_switch_timezone == 'EET'
        assert isinstance(profile, ProvisioningProfile)
        client.create_provisioning_profile.assert_called_with({'name': 'name', 'version': 4, 'brokerTimezone': 'EET',
                                                               'brokerDSTSwitchTimezone': 'EET'})

    @pytest.mark.asyncio
    async def test_reload_provisioning_profile(self):
        """Should reload provisioning profile."""

        client.get_provisioning_profile = AsyncMock(side_effect=[{'_id': 'id', 'name': 'name',
                                                                  'version': 4, 'status': 'new',
                                                                  'brokerTimezone': 'EET',
                                                                  'brokerDSTSwitchTimezone': 'EET'},
                                                                 {'_id': 'id', 'name': 'name',
                                                                  'version': 4, 'status': 'active',
                                                                  'brokerTimezone': 'EET',
                                                                  'brokerDSTSwitchTimezone': 'EET'
                                                                  }
                                                                 ])
        profile = await api.get_provisioning_profile('id')
        await profile.reload()
        assert profile.status == 'active'
        client.get_provisioning_profile.assert_called_with('id')
        assert client.get_provisioning_profile.call_count == 2

    @pytest.mark.asyncio
    async def test_remove_provisioning_profile(self):
        """Should remove provisioning profile."""

        client.get_provisioning_profile = AsyncMock(return_value={'_id': 'id', 'name': 'name',
                                                                  'version': 4, 'status': 'new',
                                                                  'brokerTimezone': 'EET',
                                                                  'brokerDSTSwitchTimezone': 'EET'
                                                                  })
        client.delete_provisioning_profile = AsyncMock()
        profile = await api.get_provisioning_profile('id')
        await profile.remove()
        client.delete_provisioning_profile.assert_called_with('id')

    @pytest.mark.asyncio
    async def test_upload_file_to_provisioning_profile(self):
        """Should upload a file to provisioning profile."""

        client.get_provisioning_profile = AsyncMock(return_value={'_id': 'id', 'name': 'name',
                                                                  'version': 4, 'status': 'new',
                                                                  'brokerTimezone': 'EET',
                                                                  'brokerDSTSwitchTimezone': 'EET'
                                                                  })
        client.upload_provisioning_profile_file = AsyncMock()
        profile = await api.get_provisioning_profile('id')
        await profile.upload_file('broker.srv', '/path/to/file.srv')
        client.upload_provisioning_profile_file.assert_called_with('id', 'broker.srv', '/path/to/file.srv')

    @pytest.mark.asyncio
    async def test_update_provisioning_profile(self):
        """Should update provisioning profile."""

        client.get_provisioning_profile = AsyncMock(return_value={'_id': 'id', 'name': 'name',
                                                                  'version': 4, 'status': 'new',
                                                                  'brokerTimezone': 'EET',
                                                                  'brokerDSTSwitchTimezone': 'EET'
                                                                  })
        client.update_provisioning_profile = AsyncMock()
        profile = await api.get_provisioning_profile('id')
        await profile.update({'name': 'name'})
        client.update_provisioning_profile.assert_called_with('id', {'name': 'name'})



from ..metaApi_client import MetaApiClient
from ...metaApi.models import date, format_date
from .copyFactory_models import CopyFactoryTransaction, CopyFactoryStrategyIdAndName, CopyFactorySubscriberOrProvider
from datetime import datetime
from typing import List
from httpx import Response


class HistoryClient(MetaApiClient):
    """metaapi.cloud CopyFactory history API (trade copying history API) client (see
    https://trading-api-v1.project-stock.agiliumlabs.cloud/swagger/#/)"""

    def __init__(self, http_client, token: str, domain: str = 'agiliumtrade.agiliumtrade.ai'):
        """Inits CopyFactory history API client instance.

        Args:
            http_client: HTTP client.
            token: Authorization token.
            domain: Domain to connect to, default is agiliumtrade.agiliumtrade.ai.
        """
        super().__init__(http_client, token, domain)
        self._host = f'https://trading-api-v1.{domain}'

    async def get_providers(self) -> 'Response[List[CopyFactorySubscriberOrProvider]]':
        """Returns list of providers providing strategies to the current user
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default/get_users_current_providers

        Returns:
            A coroutine resolving with providers found.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_providers')
        opts = {
            'url': f"{self._host}/users/current/providers",
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def get_subscribers(self) -> 'Response[List[CopyFactorySubscriberOrProvider]]':
        """Returns list of subscribers providing strategies to the current user
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default/get_users_current_subscribers

        Returns:
            A coroutine resolving with subscribers found.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_subscribers')
        opts = {
            'url': f"{self._host}/users/current/subscribers",
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def get_strategies_subscribed(self) -> 'Response[List[CopyFactoryStrategyIdAndName]]':
        """Returns list of strategies the current user is subscribed to
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default/get_users_current_strategies_subscribed

        Returns:
            A coroutine resolving with strategies found.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_strategies_subscribed')
        opts = {
            'url': f"{self._host}/users/current/strategies-subscribed",
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def get_provided_strategies(self) -> 'Response[List[CopyFactoryStrategyIdAndName]]':
        """Returns list of strategies the current user provides to other users
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default/get_users_current_provided_strategies

        Returns:
            A coroutine resolving with strategies found.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_provided_strategies')
        opts = {
            'url': f"{self._host}/users/current/provided-strategies",
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def get_provided_strategies_transactions(self, time_from: datetime, time_till: datetime,
                                                   strategy_ids: List[str] = None, subscriber_ids: List[str] = None,
                                                   offset: int = None, limit: int = None) -> \
            'Response[List[CopyFactoryTransaction]]':
        """Returns list of transactions on the strategies the current user provides to other users
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default
        /get_users_current_provided_strategies_transactions

        Args:
            time_from: Time to load transactions from.
            time_till: Time to load transactions till.
            strategy_ids: Optional list of strategy ids to filter transactions by.
            subscriber_ids: Optional list of subscribers to filter transactions by.
            offset: Pagination offset. Default value is 0.
            limit: Pagination limit. Default value is 10000.

        Returns:
            A coroutine resolving with transactions found.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_provided_strategies_transactions')
        qs = {
            'from': format_date(time_from),
            'till': format_date(time_till)
        }
        if strategy_ids:
            qs['strategyId'] = strategy_ids
        if subscriber_ids:
            qs['subscriberId'] = subscriber_ids
        if not (offset is None):
            qs['offset'] = offset
        if limit:
            qs['limit'] = limit
        opts = {
          'url': f'{self._host}/users/current/provided-strategies/transactions',
          'method': 'GET',
          'headers': {
            'auth-token': self._token
          },
          'params': qs
        }
        transactions = await self._httpClient.request(opts)
        for transaction in transactions:
            transaction['time'] = date(transaction['time'])
        return transactions

    async def get_strategies_subscribed_transactions(self, time_from: datetime, time_till: datetime,
                                                     strategy_ids: List[str] = None, provider_ids: List[str] = None,
                                                     offset: int = None, limit: int = None) -> \
            'Response[List[CopyFactoryTransaction]]':
        """Returns list of trades on the strategies the current user subscribed to
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default
        /get_users_current_strategies_subscribed_transactions

        Args:
            time_from: Time to load transactions from.
            time_till: Time to load transactions till.
            strategy_ids: Optional list of strategy ids to filter transactions by.
            provider_ids: Optional list of providers to filter transactions by.
            offset: Pagination offset. Default value is 0.
            limit: Pagination limit. Default value is 10000.

        Returns:
            A coroutine resolving with transactions found.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_strategies_subscribed_transactions')
        qs = {
            'from': format_date(time_from),
            'till': format_date(time_till)
        }
        if strategy_ids:
            qs['strategyId'] = strategy_ids
        if provider_ids:
            qs['providerId'] = provider_ids
        if not (offset is None):
            qs['offset'] = offset
        if limit:
            qs['limit'] = limit
        opts = {
          'url': f'{self._host}/users/current/strategies-subscribed/transactions',
          'method': 'GET',
          'headers': {
            'auth-token': self._token
          },
          'params': qs
        }
        transactions = await self._httpClient.request(opts)
        for transaction in transactions:
            transaction['time'] = date(transaction['time'])
        return transactions



import os
from typing import Dict, List, Optional
from typing_extensions import TypedDict
import json
import math
from datetime import datetime
import asyncio
import functools
import shutil
from copy import deepcopy
from ...metaApi.models import date


class PacketLoggerOpts(TypedDict):
    """Packet logger options."""

    fileNumberLimit: Optional[int]
    """Maximum amount of files per account. Default is 12."""
    logFileSizeInHours: Optional[float]
    """Amount of logged hours per account file. Default is 4."""
    compressSpecifications: Optional[bool]
    """Whether to compress specifications packets. Default is true."""
    compressPrices: Optional[bool]
    """Whether to compress price packets. Default is true."""


class PacketLogger:
    """A class which records packets into log files."""

    def __init__(self, opts: PacketLoggerOpts = None):
        """Inits the class.

        Args:
            opts: Packet logger options.
        """
        opts = opts or {}
        self._fileNumberLimit = opts['fileNumberLimit'] if 'fileNumberLimit' in opts else 12
        self._logFileSizeInHours = opts['logFileSizeInHours'] if 'logFileSizeInHours' in opts else 4
        self._compressSpecifications = opts['compressSpecifications'] if 'compressSpecifications' in opts else True
        self._compressPrices = opts['compressPrices'] if 'compressPrices' in opts else True
        self._previousPrices = {}
        self._lastKeepAlive = {}
        self._writeQueue = {}
        self._root = './.metaapi/logs'
        self._recordInterval: asyncio.Task or None = None
        self._deleteOldLogsInterval: asyncio.Task or None = None
        if not os.path.exists('./.metaapi'):
            os.mkdir('./.metaapi')

        if not os.path.exists(self._root):
            os.mkdir(self._root)

    def _ensure_previous_price_object(self, account_id: str):
        if account_id not in self._previousPrices:
            self._previousPrices[account_id] = {}

    def log_packet(self, packet: Dict):
        """Processes packets and pushes them into save queue.

        Args:
            packet: Packet to log.
        """
        instance_index = packet['instanceIndex'] if 'instanceIndex' in packet else 0
        packet = deepcopy(packet)
        if packet['accountId'] not in self._writeQueue:
            self._writeQueue[packet['accountId']] = {'isWriting': False, 'queue': []}
        if packet['type'] == 'status':
            return
        if packet['accountId'] not in self._lastKeepAlive:
            self._lastKeepAlive[packet['accountId']] = {}
        if packet['type'] == 'keepalive':
            self._lastKeepAlive[packet['accountId']][instance_index] = packet
            return
        queue: List = self._writeQueue[packet['accountId']]['queue']
        if packet['accountId'] not in self._previousPrices:
            self._previousPrices[packet['accountId']] = {}
        prev_price = self._previousPrices[packet['accountId']][instance_index] if \
            instance_index in self._previousPrices[packet['accountId']] else None
        if packet['type'] != 'prices':
            if prev_price is not None:
                self._record_prices(packet['accountId'], instance_index)
            if packet['type'] == 'specifications' and self._compressSpecifications:
                queue.append(json.dumps({'type': packet['type'], 'sequenceNumber': packet['sequenceNumber'] if
                                         'sequenceNumber' in packet else None, 'sequenceTimestamp':
                                         packet['sequenceTimestamp'] if 'sequenceTimestamp' in packet else None,
                                         'instanceIndex': instance_index}))
            else:
                queue.append(json.dumps(packet))
        else:
            if not self._compressPrices:
                queue.append(json.dumps(packet))
            else:
                if prev_price is not None:
                    valid_sequence_numbers = [prev_price['last']['sequenceNumber'],
                                              prev_price['last']['sequenceNumber'] + 1]
                    if instance_index in self._lastKeepAlive[packet['accountId']] and 'sequenceNumber' in \
                            self._lastKeepAlive[packet['accountId']][instance_index]:
                        valid_sequence_numbers.append(
                            self._lastKeepAlive[packet['accountId']][instance_index]['sequenceNumber'] + 1)
                    if packet['sequenceNumber'] not in valid_sequence_numbers:
                        self._record_prices(packet['accountId'], instance_index)
                        self._ensure_previous_price_object(packet['accountId'])
                        self._previousPrices[packet['accountId']][instance_index] = {'first': packet, 'last': packet}
                        queue.append(json.dumps(packet))
                    else:
                        self._previousPrices[packet['accountId']][instance_index]['last'] = packet
                else:
                    if 'sequenceNumber' in packet:
                        self._ensure_previous_price_object(packet['accountId'])
                        self._previousPrices[packet['accountId']][instance_index] = {'first': packet, 'last': packet}
                    queue.append(json.dumps(packet))

    async def read_logs(self, account_id: str, date_after: datetime = None, date_before: datetime = None):
        """Returns log messages within date bounds as an array of objects.

        Args:
            account_id: Account id.
            date_after: Date to get logs after.
            date_before: Date to get logs before.
        """
        folders = os.listdir(self._root)
        folders.sort()
        packets = []
        for folder in folders:
            file_path = f'{self._root}/{folder}/{account_id}.log'
            if os.path.exists(file_path):
                contents = open(file_path, "r").readlines()
                messages = list(map(lambda message: {'date': date(message[1:24]), 'message':
                                    message[26:].replace('\n', '')}, contents))
                if date_after:
                    messages = list(filter(lambda message: message['date'] > date_after, messages))
                if date_before:
                    messages = list(filter(lambda message: message['date'] < date_before, messages))
                packets += messages
        return packets

    def get_file_path(self, account_id) -> str:
        """Returns path for account log file.

        Args:
            account_id: Account id.

        Returns:
            File path.
        """
        file_index = math.floor(datetime.now().hour / self._logFileSizeInHours)
        folder_name = f'{datetime.now().strftime("%Y-%m-%d")}-{file_index if file_index > 9 else "0" + str(file_index)}'
        if not os.path.exists(f'{self._root}/{folder_name}'):
            os.mkdir(f'{self._root}/{folder_name}')
        return f'{self._root}/{folder_name}/{account_id}.log'

    def start(self):
        """Initializes the packet logger."""
        self._previousPrices = {}

        async def record_job():
            while True:
                await asyncio.sleep(1)
                await self._append_logs()

        async def delete_old_data_job():
            while True:
                await asyncio.sleep(10)
                await self._delete_old_data()

        if not self._recordInterval:
            self._recordInterval = asyncio.create_task(record_job())
            self._deleteOldLogsInterval = asyncio.create_task(delete_old_data_job())

    def stop(self):
        """Deinitializes the packet logger."""
        self._recordInterval.cancel()
        self._recordInterval = None
        self._deleteOldLogsInterval.cancel()
        self._deleteOldLogsInterval = None

    def _record_prices(self, account_id: str, instance_index: int):
        """Records price packet messages to log files.

        Args:
            account_id: Account id.
        """
        prev_price = self._previousPrices[account_id][instance_index] if instance_index \
            in self._previousPrices[account_id] else {'first': {}, 'last': {}}
        queue = self._writeQueue[account_id]['queue']
        del self._previousPrices[account_id][instance_index]
        if not len(self._previousPrices[account_id].keys()):
            del self._previousPrices[account_id]
        if prev_price['first']['sequenceNumber'] != prev_price['last']['sequenceNumber']:
            queue.append(json.dumps(prev_price['last']))
            queue.append(f'Recorded price packets {prev_price["first"]["sequenceNumber"]}'
                         f'-{prev_price["last"]["sequenceNumber"]}, instanceIndex: {instance_index}')

    async def _append_logs(self):
        """Writes logs to files."""
        for key in self._writeQueue:
            queue = self._writeQueue[key]
            if (not queue['isWriting']) and len(queue['queue']):
                queue['isWriting'] = True
                try:
                    file_path = self.get_file_path(key)
                    write_string = functools.reduce(
                        lambda a, b: a + f'[{datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]}] {b}\r',
                        queue['queue'], '')
                    queue['queue'] = []
                    f = open(file_path, "a+")
                    f.write(write_string)
                    f.close()
                except Exception as err:
                    print('Error writing log', err)
                queue['isWriting'] = False

    async def _delete_old_data(self):
        """Deletes folders when the folder limit is exceeded."""
        contents = os.listdir(self._root)
        contents.sort()
        for folder_name in list(reversed(contents))[self._fileNumberLimit:]:
            shutil.rmtree(f'{self._root}/{folder_name}')



from ..clients.metaApi.metatraderDemoAccount_client import MetatraderDemoAccountDto


class MetatraderDemoAccount:
    """Implements a MetaTrader demo account entity."""

    def __init__(self, data: MetatraderDemoAccountDto):
        """Inits a MetaTrader demo account entity.

        Args:
            data: MetaTrader demo account data.
        """
        self._data = data

    @property
    def login(self) -> str:
        """Returns account login.

        Returns:
            Account login.
        """
        return self._data['login']

    @property
    def password(self) -> str:
        """Returns account password.

        Returns:
            Account password.
        """
        return self._data['password']

    @property
    def server_name(self) -> str:
        """Returns account server name.

        Returns:
            Account server name.
        """
        return self._data['serverName']

    @property
    def investor_password(self) -> str:
        """Returns account investor password.

        Returns:
            Account investor password.
        """
        return self._data['investorPassword']



from typing import Callable, Dict
from .avlTreeReservoir import reservoir


class StatisticalReservoir:
    """Statistical reservoir of a fixed size capable calculating percentiles."""

    def __init__(self, size: int, interval: int, random_number_gen: Callable = None):
        """Inits reservoir.

        Args:
            size: Reservoir size.
            interval: Reservoir interval in milliseconds.
            random_number_gen: Custom number generator.
        """
        self.reservoir = reservoir(size, interval, random_number_gen)
        self.length = self.reservoir['size']()

    def push_measurement(self, data: float):
        """Add element to reservoir.

        Args:
            data: Data to add."""
        self.reservoir['pushSome'](data)
        self.length = self.reservoir['size']()

    def get_percentile(self, p: float):
        """Calculate percentile statistics for values stored in reservoir.

        Args:
            p: Value in percents from 0 to 100.

        Returns:
            Percentile value."""
        self.length = self.reservoir['size']()
        return self.reservoir['getPercentile'](p)

    def restore_values(self, value: Dict):
        """Restore reservoir from saving data.

        Args:
            value: Stored value."""
        self.reservoir['restoreValues'](value)

    def to_array(self):
        return self.reservoir['toArray']()

    def to_value_array(self):
        return self.reservoir['toValueArray']()



import pytest
from mock import patch
from .packetLogger import PacketLogger
from typing import Dict
from copy import deepcopy
from ...metaApi.models import date
from asyncio import sleep
import shutil
import json
import os
from freezegun import freeze_time
start_time = '2020-10-10 00:00:01.000'
packets = {}
packet_logger: PacketLogger = None
folder = './.metaapi/logs/'


def change_sn(obj: Dict, sequence_number: int, instance_index: int = 7):
    new_obj = deepcopy(obj)
    new_obj['sequenceNumber'] = sequence_number
    new_obj['instanceIndex'] = instance_index
    return new_obj


@pytest.fixture(autouse=True)
async def run_around_tests():
    with patch('lib.clients.metaApi.packetLogger.asyncio.sleep', new=lambda x: sleep(x / 50)):
        global packet_logger
        packet_logger = PacketLogger({'fileNumberLimit': 3, 'logFileSizeInHours': 4})
        packet_logger.start()
        global packets
        packets = {
            'accountInformation': {
                'type': 'accountInformation',
                'instanceIndex': 7,
                'accountInformation': {
                    'broker': 'Broker',
                    'currency': 'USD',
                    'server': 'Broker-Demo',
                    'balance': 20000,
                    'equity': 25000
                },
                'accountId': 'accountId',
                'sequenceTimestamp': 100000,
            },
            'prices': {
                'type': 'prices',
                'instanceIndex': 7,
                'prices': [{
                    'symbol': 'EURUSD',
                    'bid': 1.18,
                    'ask': 1.19
                },
                    {
                        'symbol': 'USDJPY',
                        'bid': 103.222,
                        'ask': 103.25
                    }],
                'accountId': 'accountId',
                'sequenceNumber': 1,
                'sequenceTimestamp': 100000,
            },
            'status': {
                'status': 'connected',
                'instanceIndex': 7,
                'type': 'status',
                'accountId': 'accountId',
                'sequenceTimestamp': 100000,
            },
            'keepalive': {
                'instanceIndex': 7,
                'type': 'keepalive',
                'accountId': 'accountId',
                'sequenceTimestamp': 100000
            },
            'specifications': {
                'specifications': [],
                'instanceIndex': 7,
                'type': 'specifications',
                'accountId': 'accountId',
                'sequenceTimestamp': 100000,
                'sequenceNumber': 1
            }
        }
        yield
        packet_logger.stop()
        shutil.rmtree(folder)


class TestPacketLogger:
    @pytest.mark.asyncio
    async def test_record_packet(self):
        """Should record packet."""
        packet_logger.log_packet(packets['accountInformation'])
        await sleep(0.04)
        result = await packet_logger.read_logs('accountId')
        assert json.loads(result[0]['message']) == packets['accountInformation']

    @pytest.mark.asyncio
    async def test_record_price_packets_without_sn(self):
        """Should record price packets without sequence number."""
        packet = packets['prices']
        del packet['sequenceNumber']
        packet_logger.log_packet(packet)
        packet_logger.log_packet(packets['prices'])
        packet_logger.log_packet(packet)
        await sleep(0.04)
        result = await packet_logger.read_logs('accountId')
        assert json.loads(result[0]['message']) == packet

    @pytest.mark.asyncio
    async def test_not_record_status(self):
        """Should not record status and keepalive packets."""
        packet_logger.log_packet(packets['status'])
        packet_logger.log_packet(packets['keepalive'])
        await sleep(0.04)
        result = await packet_logger.read_logs('accountId')
        assert len(result) == 0

    @pytest.mark.asyncio
    async def test_record_short_specifications(self):
        """Should record shortened specifications."""
        packet_logger.log_packet(packets['specifications'])
        await sleep(0.04)
        result = await packet_logger.read_logs('accountId')
        assert json.loads(result[0]['message']) == {'type': 'specifications', 'sequenceNumber': 1,
                                                    'sequenceTimestamp': 100000, 'instanceIndex': 7}

    @pytest.mark.asyncio
    async def test_record_full_specifications(self):
        """Should record full specifications if compress disabled."""
        global packet_logger
        packet_logger.stop()
        packet_logger = PacketLogger({'compressSpecifications': False})
        packet_logger.start()
        packet_logger.log_packet(packets['specifications'])
        await sleep(0.04)
        result = await packet_logger.read_logs('accountId')
        assert json.loads(result[0]['message']) == packets['specifications']

    @pytest.mark.asyncio
    async def test_record_single_price_packet(self):
        """Should record single price packet."""
        packet_logger.log_packet(packets['prices'])
        packet_logger.log_packet(packets['accountInformation'])
        await sleep(0.04)
        result = await packet_logger.read_logs('accountId')
        assert json.loads(result[0]['message']) == packets['prices']
        assert json.loads(result[1]['message']) == packets['accountInformation']

    @pytest.mark.asyncio
    async def test_record_range_of_price_packets(self):
        """Should record range of price packets."""
        packet_logger.log_packet(packets['prices'])
        packet_logger.log_packet(change_sn(packets['prices'], 2))
        packet_logger.log_packet(change_sn(packets['prices'], 3))
        packet_logger.log_packet(change_sn(packets['prices'], 4))
        packet_logger.log_packet(change_sn(packets['keepalive'], 5))
        packet_logger.log_packet(change_sn(packets['prices'], 6))
        packet_logger.log_packet(packets['accountInformation'])
        await sleep(0.04)
        result = await packet_logger.read_logs('accountId')
        assert json.loads(result[0]['message']) == packets['prices']
        assert json.loads(result[1]['message']) == change_sn(packets['prices'], 6)
        assert result[2]['message'] == 'Recorded price packets 1-6, instanceIndex: 7'
        assert json.loads(result[3]['message']) == packets['accountInformation']

    @pytest.mark.asyncio
    async def test_record_range_of_price_packets_different_instances(self):
        """Should record range of price packets of different instances."""
        packet_logger.log_packet(packets['prices'])
        packet_logger.log_packet(change_sn(packets['prices'], 2))
        packet_logger.log_packet(change_sn(packets['prices'], 3))
        packet_logger.log_packet(change_sn(packets['prices'], 1, 8))
        packet_logger.log_packet(change_sn(packets['prices'], 2, 8))
        packet_logger.log_packet(change_sn(packets['prices'], 3, 8))
        packet_logger.log_packet(change_sn(packets['prices'], 4, 8))
        packet_logger.log_packet(change_sn(packets['prices'], 4))
        packet_logger.log_packet(change_sn(packets['prices'], 5, 8))
        account_info = deepcopy(packets['accountInformation'])
        account_info['instanceIndex'] = 8
        packet_logger.log_packet(account_info)
        packet_logger.log_packet(change_sn(packets['prices'], 5))
        packet_logger.log_packet(packets['accountInformation'])
        await sleep(0.04)
        result = await packet_logger.read_logs('accountId')
        assert json.loads(result[0]['message']) == packets['prices']
        assert json.loads(result[1]['message']) == change_sn(packets['prices'], 1, 8)
        assert json.loads(result[2]['message']) == change_sn(packets['prices'], 5, 8)
        assert result[3]['message'] == 'Recorded price packets 1-5, instanceIndex: 8'
        assert json.loads(result[4]['message']) == account_info
        assert json.loads(result[5]['message']) == change_sn(packets['prices'], 5)
        assert result[6]['message'] == 'Recorded price packets 1-5, instanceIndex: 7'
        assert json.loads(result[7]['message']) == packets['accountInformation']

    @pytest.mark.asyncio
    async def test_record_all_price_packets(self):
        """Should record all price packets if compress disabled."""
        global packet_logger
        packet_logger.stop()
        packet_logger = PacketLogger({'compressPrices': False})
        packet_logger.start()
        packet_logger.log_packet(packets['prices'])
        packet_logger.log_packet(change_sn(packets['prices'], 2))
        packet_logger.log_packet(change_sn(packets['prices'], 3))
        packet_logger.log_packet(change_sn(packets['prices'], 4))
        packet_logger.log_packet(change_sn(packets['prices'], 5))
        packet_logger.log_packet(packets['accountInformation'])
        await sleep(0.04)
        result = await packet_logger.read_logs('accountId')
        assert json.loads(result[0]['message']) == packets['prices']
        assert json.loads(result[1]['message']) == change_sn(packets['prices'], 2)
        assert json.loads(result[2]['message']) == change_sn(packets['prices'], 3)
        assert json.loads(result[3]['message']) == change_sn(packets['prices'], 4)
        assert json.loads(result[4]['message']) == change_sn(packets['prices'], 5)
        assert json.loads(result[5]['message']) == packets['accountInformation']

    @pytest.mark.asyncio
    async def test_stop_price_if_sn_doesnt_match(self):
        """Should stop price packet sequence if price sequence number doesnt match."""
        packet_logger.log_packet(packets['prices'])
        packet_logger.log_packet(change_sn(packets['prices'], 2))
        packet_logger.log_packet(change_sn(packets['prices'], 3))
        packet_logger.log_packet(change_sn(packets['prices'], 4))
        packet_logger.log_packet(change_sn(packets['prices'], 6))
        await sleep(0.04)
        result = await packet_logger.read_logs('accountId')
        assert json.loads(result[0]['message']) == packets['prices']
        assert json.loads(result[1]['message']) == change_sn(packets['prices'], 4)
        assert result[2]['message'] == 'Recorded price packets 1-4, instanceIndex: 7'
        assert json.loads(result[3]['message']) == change_sn(packets['prices'], 6)

    @pytest.mark.asyncio
    async def test_read_logs_within_bounds(self):
        """Should read logs within bounds."""
        with freeze_time(start_time) as frozen_datetime:
            packet_logger.log_packet(packets['accountInformation'])
            packet_logger.log_packet(packets['accountInformation'])
            await sleep(0.04)
            frozen_datetime.move_to('2020-10-10 01:00:01.000')
            packet_logger.log_packet(packets['accountInformation'])
            packet_logger.log_packet(packets['accountInformation'])
            packet_logger.log_packet(packets['accountInformation'])
            packet_logger.log_packet(packets['accountInformation'])
            packet_logger.log_packet(packets['accountInformation'])
            await sleep(0.04)
            frozen_datetime.move_to('2020-10-10 01:30:01.000')
            packet_logger.log_packet(packets['accountInformation'])
            packet_logger.log_packet(packets['accountInformation'])
            packet_logger.log_packet(packets['accountInformation'])
            await sleep(0.04)
            result = await packet_logger.read_logs('accountId', date('2020-10-10 00:30:00.000'),
                                                   date('2020-10-10 01:30:00.000'))
            assert len(result) == 5
            result_after = await packet_logger.read_logs('accountId', date('2020-10-10 00:30:00.000'))
            assert len(result_after) == 8
            result_before = await packet_logger.read_logs('accountId', None, date('2020-10-10 01:30:00.000'))
            assert len(result_before) == 7

    @pytest.mark.asyncio
    async def test_delete_expired_folders(self):
        """Should delete expired folders."""
        with freeze_time(start_time) as frozen_datetime:
            packet_logger.log_packet(packets['accountInformation'])
            await sleep(0.21)
            folder_list = os.listdir(folder)
            assert folder_list == ['2020-10-10-00']

            frozen_datetime.move_to('2020-10-10 05:00:01.000')
            packet_logger.log_packet(packets['accountInformation'])
            await sleep(0.21)
            folder_list = os.listdir(folder)
            folder_list.sort()
            assert os.listdir(folder) == ['2020-10-10-00', '2020-10-10-01']

            frozen_datetime.move_to('2020-10-10 09:00:01.000')
            packet_logger.log_packet(packets['accountInformation'])
            await sleep(0.21)
            folder_list = os.listdir(folder)
            folder_list.sort()
            assert folder_list == ['2020-10-10-00', '2020-10-10-01', '2020-10-10-02']

            frozen_datetime.move_to('2020-10-10 13:00:01.000')
            packet_logger.log_packet(packets['accountInformation'])
            await sleep(0.21)
            folder_list = os.listdir(folder)
            folder_list.sort()
            assert folder_list == ['2020-10-10-01', '2020-10-10-02', '2020-10-10-03']
